import copy
import uuid
from datetime import datetime

import pymongo
from google.protobuf.json_format import MessageToDict

from fedn.network.state import ReducerStateToString, StringToReducerState


class MongoStateStore:
    """Statestore implementation using MongoDB.

    :param network_id: The network id.
    :type network_id: str
    :param config: The statestore configuration.
    :type config: dict
    :param defaults: The default configuration. Given by config/settings-reducer.yaml.template
    :type defaults: dict
    """

    def __init__(self, network_id, config):
        """Constructor."""
        self.__inited = False
        try:
            self.config = config
            self.network_id = network_id
            self.mdb = self.connect()

            # FEDn network
            self.network = self.mdb["network"]
            self.reducer = self.network["reducer"]
            self.combiners = self.network["combiners"]
            self.clients = self.network["clients"]
            self.storage = self.network["storage"]

            # Control
            self.control = self.mdb["control"]
            self.package = self.control["package"]
            self.state = self.control["state"]
            self.model = self.control["model"]
            self.sessions = self.control["sessions"]
            self.rounds = self.control["rounds"]
            self.validations = self.control["validations"]

            # Logging
            self.status = self.control["status"]

            self.__inited = True
        except Exception as e:
            print("FAILED TO CONNECT TO MONGODB, {}".format(e), flush=True)
            self.state = None
            self.model = None
            self.control = None
            self.network = None
            self.combiners = None
            self.clients = None
            raise

        self.init_index()

    def connect(self):
        """ Establish client connection to MongoDB.

        :param config: Dictionary containing connection strings and security credentials.
        :type config: dict
        :param network_id: Unique identifier for the FEDn network, used as db name
        :type network_id: str
        :return: MongoDB client pointing to the db corresponding to network_id
        """
        try:
            mc = pymongo.MongoClient(**self.config)
            # This is so that we check that the connection is live
            mc.server_info()
            mdb = mc[self.network_id]
            return mdb
        except Exception:
            raise

    def init_index(self):
        self.package.create_index([("id", pymongo.DESCENDING)])

    def is_inited(self):
        """Check if the statestore is intialized.

        :return: True if initialized, else False.
        :rtype: bool
        """
        return self.__inited

    def get_config(self):
        """Retrive the statestore config.

        :return: The statestore config.
        :rtype: dict
        """
        data = {
            "type": "MongoDB",
            "mongo_config": self.config,
            "network_id": self.network_id,
        }
        return data

    def state(self):
        """Get the current state.

        :return: The current state.
        :rtype: str
        """
        return StringToReducerState(self.state.find_one()["current_state"])

    def transition(self, state):
        """Transition to a new state.

        :param state: The new state.
        :type state: str
        :return:
        """
        old_state = self.state.find_one({"state": "current_state"})
        if old_state != state:
            return self.state.update_one(
                {"state": "current_state"},
                {"$set": {"state": ReducerStateToString(state)}},
                True,
            )
        else:
            print(
                "Not updating state, already in {}".format(
                    ReducerStateToString(state)
                )
            )

    def get_sessions(self, limit=None, skip=None, sort_key="_id", sort_order=pymongo.DESCENDING):
        """Get all sessions.

        :param limit: The maximum number of sessions to return.
        :type limit: int
        :param skip: The number of sessions to skip.
        :type skip: int
        :param sort_key: The key to sort by.
        :type sort_key: str
        :param sort_order: The sort order.
        :type sort_order: pymongo.ASCENDING or pymongo.DESCENDING
        :return: Dictionary of sessions in result (array of session objects) and count.
        """

        result = None

        if limit is not None and skip is not None:
            limit = int(limit)
            skip = int(skip)

            result = self.sessions.find().limit(limit).skip(skip).sort(
                sort_key, sort_order
            )
        else:
            result = self.sessions.find().sort(
                sort_key, sort_order
            )

        count = self.sessions.count_documents({})

        return {
            "result": result,
            "count": count,
        }

    def get_session(self, session_id):
        """Get session with id.

        :param session_id: The session id.
        :type session_id: str
        :return: The session.
        :rtype: ObjectID
        """
        return self.sessions.find_one({"session_id": session_id})

    def set_latest_model(self, model_id, session_id=None):
        """Set the latest model id.

        :param model_id: The model id.
        :type model_id: str
        :return:
        """

        committed_at = datetime.now()
        current_model = self.model.find_one({"key": "current_model"})
        parent_model = None

        # if session_id is set the it means the model is generated from a session
        # and we need to set the parent model
        # if not the model is uploaded by the user and we don't need to set the parent model
        if session_id is not None:
            parent_model = current_model["model"] if current_model and "model" in current_model else None

        self.model.insert_one(
            {
                "key": "models",
                "model": model_id,
                "parent_model": parent_model,
                "session_id": session_id,
                "committed_at": committed_at,
            }
        )

        self.model.update_one(
            {"key": "current_model"}, {"$set": {"model": model_id}}, True
        )
        self.model.update_one(
            {"key": "model_trail"},
            {
                "$push": {
                    "model": model_id,
                    "committed_at": str(committed_at),
                }
            },
            True,
        )

    def get_initial_model(self):
        """Return model_id for the initial model in the model trail

        :return: The initial model id. None if no model is found.
        :rtype: str
        """

        result = self.model.find_one(
            {"key": "model_trail"}, sort=[("committed_at", pymongo.ASCENDING)]
        )
        if result is None:
            return None

        try:
            model_id = result["model"]
            if model_id == "" or model_id == " ":
                return None
            return model_id[0]
        except (KeyError, IndexError):
            return None

    def get_latest_model(self):
        """Return model_id for the latest model in the model_trail

        :return: The latest model id. None if no model is found.
        :rtype: str
        """
        result = self.model.find_one({"key": "current_model"})
        if result is None:
            return None

        try:
            model_id = result["model"]
            if model_id == "" or model_id == " ":
                return None
            return model_id
        except (KeyError, IndexError):
            return None

    def set_current_model(self, model_id: str):
        """Set the current model in statestore.

        :param model_id: The model id.
        :type model_id: str
        :return:
        """

        try:

            committed_at = datetime.now()

            existing_model = self.model.find_one({"key": "models", "model": model_id})

            if existing_model is not None:

                self.model.update_one(
                    {"key": "current_model"}, {"$set": {"model": model_id, "committed_at": committed_at, "session_id": None}}, True
                )

                return True
        except Exception as e:
            print("ERROR: {}".format(e), flush=True)

        return False

    def get_latest_round(self):
        """Get the id of the most recent round.

        :return: The id of the most recent round.
        :rtype: ObjectId
        """

        return self.rounds.find_one(sort=[("_id", pymongo.DESCENDING)])

    def get_round(self, id):
        """Get round with id.

        :param id: id of round to get
        :type id: int
        :return: round with id, reducer and combiners
        :rtype: ObjectId
        """

        return self.rounds.find_one({"round_id": str(id)})

    def get_rounds(self):
        """Get all rounds.

        :return: All rounds.
        :rtype: ObjectId
        """

        return self.rounds.find()

    def get_validations(self, **kwargs):
        """Get validations from the database.

        :param kwargs: query to filter validations
        :type kwargs: dict
        :return: validations matching query
        :rtype: ObjectId
        """

        result = self.control.validations.find(kwargs)
        return result

    def set_active_compute_package(self, id: str):
        """Set the active compute package in statestore.

        :param id: The id of the compute package (not document _id).
        :type id: str
        :return: True if successful.
        :rtype: bool
        """

        try:

            find = {"id": id}
            projection = {"_id": False, "key": False}

            doc = self.control.package.find_one(find, projection)

            if doc is None:
                return False

            doc["key"] = "active"

            self.control.package.replace_one(
                {"key": "active"}, doc
            )

        except Exception as e:
            print("ERROR: {}".format(e), flush=True)
            return False

        return True

    def set_compute_package(self, file_name: str, storage_file_name: str, helper_type: str, name: str = None, description: str = None):
        """Set the active compute package in statestore.

        :param file_name: The file_name of the compute package.
        :type file_name: str
        :return: True if successful.
        :rtype: bool
        """

        obj = {
            "file_name": file_name,
            "storage_file_name": storage_file_name,
            "helper": helper_type,
            "committed_at": datetime.now(),
            "name": name,
            "description": description,
            "id": str(uuid.uuid4()),
        }

        self.control.package.update_one(
            {"key": "active"},
            {
                "$set": obj
            },
            True,
        )

        trail_obj = {**{"key": "package_trail"}, **obj}

        self.control.package.insert_one(trail_obj)

        return True

    def get_compute_package(self):
        """Get the active compute package.

        :return: The active compute package.
        :rtype: ObjectID
        """
        try:

            find = {"key": "active"}
            projection = {"key": False, "_id": False}
            ret = self.control.package.find_one(find, projection)
            return ret
        except Exception as e:
            print("ERROR: {}".format(e), flush=True)
            return None

    def list_compute_packages(self, limit: int = None, skip: int = None, sort_key="committed_at", sort_order=pymongo.DESCENDING):
        """List compute packages in the statestore (paginated).

        :param limit: The maximum number of compute packages to return.
        :type limit: int
        :param skip: The number of compute packages to skip.
        :type skip: int
        :param sort_key: The key to sort by.
        :type sort_key: str
        :param sort_order: The sort order.
        :type sort_order: pymongo.ASCENDING or pymongo.DESCENDING
        :return: Dictionary of compute packages in result and count.
        :rtype: dict
        """

        result = None
        count = None

        find_option = {"key": "package_trail"}
        projection = {"key": False, "_id": False}

        try:
            if limit is not None and skip is not None:
                result = self.control.package.find(find_option, projection).limit(limit).skip(skip).sort(sort_key, sort_order)
            else:
                result = self.control.package.find(find_option, projection).sort(sort_key, sort_order)

            count = self.control.package.count_documents(find_option)

        except Exception as e:
            print("ERROR: {}".format(e), flush=True)
            return None

        return {
            "result": result or [],
            "count": count or 0,
        }

    def set_helper(self, helper):
        """Set the active helper package in statestore.

        :param helper: The name of the helper package. See helper.py for available helpers.
        :type helper: str
        :return:
        """
        self.control.package.update_one(
            {"key": "active"}, {"$set": {"helper": helper}}, True
        )

    def get_helper(self):
        """Get the active helper package.

        :return: The active helper set for the package.
        :rtype: str
        """
        ret = self.control.package.find_one({"key": "active"})
        # if local compute package used, then 'package' is None
        # if not ret:
        # get framework from round_config instead
        #    ret = self.control.config.find_one({'key': 'round_config'})
        try:
            retcheck = ret["helper"]
            if (
                retcheck == "" or retcheck == " "
            ):  # ugly check for empty string
                return None
            return retcheck
        except (KeyError, IndexError):
            return None

    def list_models(
        self,
        session_id=None,
        limit=None,
        skip=None,
        sort_key="committed_at",
        sort_order=pymongo.DESCENDING,
    ):
        """List all models in the statestore.

        :param session_id: The session id.
        :type session_id: str
        :param limit: The maximum number of models to return.
        :type limit: int
        :param skip: The number of models to skip.
        :type skip: int
        :return: List of models.
        :rtype: list
        """
        result = None

        find_option = (
            {"key": "models"}
            if session_id is None
            else {"key": "models", "session_id": session_id}
        )

        projection = {"_id": False, "key": False}

        if limit is not None and skip is not None:
            limit = int(limit)
            skip = int(skip)

            result = (
                self.model.find(find_option, projection)
                .limit(limit)
                .skip(skip)
                .sort(sort_key, sort_order)
            )

        else:
            result = self.model.find(find_option, projection).sort(
                sort_key, sort_order
            )

        count = self.model.count_documents(find_option)

        return {
            "result": result,
            "count": count,
        }

    def get_model_trail(self):
        """Get the model trail.

        :return: dictionary of model_id: committed_at
        :rtype: dict
        """
        # TODO Make it so that model order from db is preserved.
        result = self.model.find_one({"key": "model_trail"})
        try:
            if result is not None:
                committed_at = result["committed_at"]
                model = result["model"]
                model_dictionary = dict(zip(model, committed_at))
                return model_dictionary
            else:
                return None
        except (KeyError, IndexError):
            return None

    def get_model_ancestors(self, model_id: str, limit: int):
        """Get the model ancestors.

        :param model_id: The model id.
        :type model_id: str
        :param limit: The maximum number of ancestors to return.
        :type limit: int
        :return: List of model ancestors.
        :rtype: list
        """
        model = self.model.find_one({"key": "models", "model": model_id})
        current_model_id = model["parent_model"] if model is not None else None
        result = []

        for _ in range(limit):
            if current_model_id is None:
                break

            model = self.model.find_one({"key": "models", "model": current_model_id})

            if model is not None:
                result.append(model)
                current_model_id = model["parent_model"]

        return result

    def get_model_descendants(self, model_id: str, limit: int):
        """Get the model descendants.

        :param model_id: The model id.
        :type model_id: str
        :param limit: The maximum number of descendants to return.
        :type limit: int
        :return: List of model descendants.
        :rtype: list
        """

        model: object = self.model.find_one({"key": "models", "model": model_id})
        current_model_id: str = model["model"] if model is not None else None
        result: list = []

        for _ in range(limit):
            if current_model_id is None:
                break

            model: str = self.model.find_one({"key": "models", "parent_model": current_model_id})

            if model is not None:
                result.append(model)
                current_model_id = model["model"]

        result.reverse()

        return result

    def get_model(self, model_id):
        """Get model with id.

        :param model_id: id of model to get
        :type model_id: str
        :return: model with id
        :rtype: ObjectId
        """
        return self.model.find_one({"key": "models", "model": model_id})

    def get_events(self, **kwargs):
        """Get events from the database.

        :param kwargs: query to filter events
        :type kwargs: dict
        :return: events matching query
        :rtype: ObjectId
        """
        # check if kwargs is empty

        result = None
        count = None
        projection = {"_id": False}

        if not kwargs:
            result = self.control.status.find({}, projection).sort(
                "timestamp", pymongo.DESCENDING
            )
            count = self.control.status.count_documents({})
        else:
            limit = kwargs.pop("limit", None)
            skip = kwargs.pop("skip", None)

            if limit is not None and skip is not None:
                limit = int(limit)
                skip = int(skip)
                result = (
                    self.control.status.find(kwargs, projection)
                    .sort("timestamp", pymongo.DESCENDING)
                    .limit(limit)
                    .skip(skip)
                )
            else:
                result = self.control.status.find(kwargs, projection).sort(
                    "timestamp", pymongo.DESCENDING
                )

            count = self.control.status.count_documents(kwargs)

        return {
            "result": result,
            "count": count,
        }

    def get_storage_backend(self):
        """Get the storage backend.

        :return: The storage backend.
        :rtype: ObjectID
        """
        try:
            ret = self.storage.find(
                {"status": "enabled"}, projection={"_id": False}
            )
            return ret[0]
        except (KeyError, IndexError):
            return None

    def set_storage_backend(self, config):
        """Set the storage backend.

        :param config: The storage backend configuration.
        :type config: dict
        :return:
        """
        config = copy.deepcopy(config)
        config["updated_at"] = str(datetime.now())
        config["status"] = "enabled"
        self.storage.update_one(
            {"storage_type": config["storage_type"]}, {"$set": config}, True
        )

    def set_reducer(self, reducer_data):
        """Set the reducer in the statestore.

        :param reducer_data: dictionary of reducer config.
        :type reducer_data: dict
        :return:
        """
        reducer_data["updated_at"] = str(datetime.now())
        self.reducer.update_one(
            {"name": reducer_data["name"]}, {"$set": reducer_data}, True
        )

    def get_reducer(self):
        """Get reducer.config.

        return: reducer config.
        rtype: ObjectId
        """
        try:
            ret = self.reducer.find_one()
            return ret
        except Exception:
            return None

    def get_combiner(self, name):
        """Get combiner by name.

        :param name: name of combiner to get.
        :type name: str
        :return: The combiner.
        :rtype: ObjectId
        """
        try:
            ret = self.combiners.find_one({"name": name})
            return ret
        except Exception:
            return None

    def get_combiners(self, limit=None, skip=None, sort_key="updated_at", sort_order=pymongo.DESCENDING, projection={}):
        """Get all combiners.

        :param limit: The maximum number of combiners to return.
        :type limit: int
        :param skip: The number of combiners to skip.
        :type skip: int
        :param sort_key: The key to sort by.
        :type sort_key: str
        :param sort_order: The sort order.
        :type sort_order: pymongo.ASCENDING or pymongo.DESCENDING
        :param projection: The projection.
        :type projection: dict
        :return: Dictionary of combiners in result and count.
        :rtype: dict
        """

        result = None
        count = None

        try:
            if limit is not None and skip is not None:
                limit = int(limit)
                skip = int(skip)
                result = self.combiners.find({}, projection).limit(limit).skip(skip).sort(sort_key, sort_order)
            else:
                result = self.combiners.find({}, projection).sort(sort_key, sort_order)

            count = self.combiners.count_documents({})

        except Exception:
            return None

        return {
            "result": result,
            "count": count,
        }

    def set_combiner(self, combiner_data):
        """Set combiner in statestore.

        :param combiner_data: dictionary of combiner config
        :type combiner_data: dict
        :return:
        """

        combiner_data["updated_at"] = str(datetime.now())
        self.combiners.update_one(
            {"name": combiner_data["name"]}, {"$set": combiner_data}, True
        )

    def delete_combiner(self, combiner):
        """Delete a combiner from statestore.

        :param combiner: name of combiner to delete.
        :type combiner: str
        :return:
        """
        try:
            self.combiners.delete_one({"name": combiner})
        except Exception:
            print(
                "WARNING, failed to delete combiner: {}".format(combiner),
                flush=True,
            )

    def set_client(self, client_data):
        """Set client in statestore.

        :param client_data: dictionary of client config.
        :type client_data: dict
        :return:
        """
        client_data["updated_at"] = str(datetime.now())
        self.clients.update_one(
            {"name": client_data["name"]}, {"$set": client_data}, True
        )

    def get_client(self, name):
        """Get client by name.

        :param name: name of client to get.
        :type name: str
        :return: The client. None if not found.
        :rtype: ObjectId
        """
        try:
            ret = self.clients.find({"key": name})
            if list(ret) == []:
                return None
            else:
                return ret
        except Exception:
            return None

    def list_clients(self, limit=None, skip=None, status=None, sort_key="last_seen", sort_order=pymongo.DESCENDING):
        """List all clients registered on the network.

        :param limit: The maximum number of clients to return.
        :type limit: int
        :param skip: The number of clients to skip.
        :type skip: int
        :param status:  online | offline
        :type status: str
        :param sort_key: The key to sort by.
        """

        result = None
        count = None

        try:
            find = {} if status is None else {"status": status}
            projection = {"_id": False, "updated_at": False}

            if limit is not None and skip is not None:
                limit = int(limit)
                skip = int(skip)
                result = self.clients.find(find, projection).limit(limit).skip(skip).sort(sort_key, sort_order)
            else:
                result = self.clients.find(find, projection).sort(sort_key, sort_order)

            count = self.clients.count_documents(find)

        except Exception as e:
            print("ERROR: {}".format(e), flush=True)

        return {
            "result": result,
            "count": count,
        }

    def list_combiners_data(self, combiners, sort_key="count", sort_order=pymongo.DESCENDING):
        """List all combiner data.

        :param combiners: list of combiners to get data for.
        :type combiners: list
        :param sort_key: The key to sort by.
        :type sort_key: str
        :param sort_order: The sort order.
        :type sort_order: pymongo.ASCENDING or pymongo.DESCENDING
        :return: list of combiner data.
        :rtype: list(ObjectId)
        """

        result = None

        try:

            pipeline = [
                {"$match": {"combiner": {"$in": combiners}, "status": "online"}},
                {"$group": {"_id": "$combiner", "count": {"$sum": 1}}},
                {"$sort": {sort_key: sort_order, "_id": pymongo.ASCENDING}}
            ] if combiners is not None else [
                {"$group": {"_id": "$combiner", "count": {"$sum": 1}}},
                {"$sort": {sort_key: sort_order, "_id": pymongo.ASCENDING}}
            ]

            result = self.clients.aggregate(pipeline)

        except Exception as e:
            print("ERROR: {}".format(e), flush=True)

        return result

    def report_status(self, msg):
        """Write status message to the database.

        :param msg: The status message.
        :type msg: str
        """
        data = MessageToDict(msg, including_default_value_fields=True)

        if self.status is not None:
            self.status.insert_one(data)

    def report_validation(self, validation):
        """Write model validation to database.

        :param validation: The model validation.
        :type validation: dict
        """
        data = MessageToDict(validation, including_default_value_fields=True)

        if self.validations is not None:
            self.validations.insert_one(data)

    def drop_status(self):
        """Drop the status collection."""
        if self.status:
            self.status.drop()

    def create_session(self, id=None):
        """ Create a new session object.

        :param id: The ID of the created session.
        :type id: uuid, str

        """
        if not id:
            id = uuid.uuid4()
        data = {'session_id': str(id)}
        self.sessions.insert_one(data)

    def create_round(self, round_data):
        """ Create a new round.

        :param round_data: Dictionary with round data.
        :type round_data: dict
        """
        # TODO: Add check if round_id already exists
        self.rounds.insert_one(round_data)

    def set_session_config(self, id, config):
        """Set the session configuration.

        :param id: The session id
        :type id: str
        :param config: Session configuration
        :type config: dict
        """
        self.sessions.update_one({'session_id': str(id)}, {
            '$push': {'session_config': config}}, True)

    def set_session_status(self, id, status):
        """Set session status.

        :param round_id: The round unique identifier
        :type round_id: str
        :param round_status: The status of the session.
        """
        self.sessions.update_one({'session_id': str(id)}, {
            '$set': {'status': status}}, True)

    def set_round_combiner_data(self, data):
        """Set combiner round controller data.

        :param data: The combiner data
        :type data: dict
        """
        self.rounds.update_one({'round_id': str(data['round_id'])}, {
            '$push': {'combiners': data}}, True)

    def set_round_config(self, round_id, round_config):
        """Set round configuration.

        :param round_id: The round unique identifier
        :type round_id: str
        :param round_config: The round configuration
        :type round_config: dict
        """
        self.rounds.update_one({'round_id': round_id}, {
            '$set': {'round_config': round_config}}, True)

    def set_round_status(self, round_id, round_status):
        """Set round status.

        :param round_id: The round unique identifier
        :type round_id: str
        :param round_status: The status of the round.
        """
        self.rounds.update_one({'round_id': round_id}, {
            '$set': {'status': round_status}}, True)

    def set_round_data(self, round_id, round_data):
        """Update round metadata

        :param round_id: The round unique identifier
        :type round_id: str
        :param round_data: The round metadata
        :type round_data: dict
        """
        self.rounds.update_one({'round_id': round_id}, {
            '$set': {'round_data': round_data}}, True)

    def update_client_status(self, clients, status):
        """ Update client status in statestore.
        :param client_name: The client name
        :type client_name: str
        :param status: The client status
        :type status: str
        :return: None
        """
        datetime_now = datetime.now()
        filter_query = {"name": {"$in": clients}}

        update_query = {"$set": {"last_seen": datetime_now, "status": status}}
        self.clients.update_many(filter_query, update_query)



import unittest

from click.testing import CliRunner
from run_cmd import check_helper_config_file


class TestReducerCLI(unittest.TestCase):

    def setUp(self):
        self.runner = CliRunner()
        self.INIT_FILE_REDUCER = {
            "network_id": "fedn-test-network",
            "token": "fedn_token",
            "control": {
                "state": "idle",
                "helper": "keras",
            },
            "statestore": {
                "type": "MongoDB",
                "mongo_config": {
                    "username": "fedn_admin",
                    "password": "password",
                    "host": "mongo",
                    "port": "6534"
                }
            },
            "storage": {
                "storage_type": "S3",
                "storage_config": {
                    "storage_hostname": "minio",
                    "storage_port": "9000",
                    "storage_access_key": "fedn_admin",
                    "storage_secret_key": "password",
                    "storage_bucket": "fedn-models",
                    "context_bucket": "fedn-context",
                    "storage_secure_mode": "False"
                }
            }
        }

    @unittest.skip
    def test_get_statestore_config_from_file(self):
        pass

    # def test_reducer_cmd_remote(self):

    #     with self.runner.isolated_filesystem():

    #         COPY_INIT_FILE = self.INIT_FILE_REDUCER
    #         del COPY_INIT_FILE["control"]["helper"]

    #         with open('settings.yaml', 'w') as f:
    #             f.write(yaml.dump(COPY_INIT_FILE))

    #         result = self.runner.invoke(reducer_cmd, ['--remote', False, '--init',"settings.yaml"])
    #         self.assertEqual(result.output, "--remote was set to False, but no helper was found in --init settings file: settings.yaml\n")
    #         self.assertEqual(result.exit_code, -1)

    def test_check_helper_config_file(self):

        self.assertEqual(check_helper_config_file(
            self.INIT_FILE_REDUCER), "keras")

        COPY_INIT_FILE = self.INIT_FILE_REDUCER
        del COPY_INIT_FILE["control"]["helper"]

        with self.assertRaises(SystemExit):
            check_helper_config_file(COPY_INIT_FILE)


if __name__ == '__main__':
    unittest.main()



""" The models package handles storing of model updates durign the federated training process.
The functionality is used by the combiner service during aggregation of model updates from clients.
By implementing the interface in the base class modelstorage.py, a developer may customize the
behaviour of the framework. """



import copy
import datetime
import time
import uuid

from tenacity import (retry, retry_if_exception_type, stop_after_delay,
                      wait_random)

from fedn.common.log_config import logger
from fedn.network.combiner.interfaces import CombinerUnavailableError
from fedn.network.combiner.modelservice import load_model_from_BytesIO
from fedn.network.controller.controlbase import ControlBase
from fedn.network.state import ReducerState


class UnsupportedStorageBackend(Exception):
    """Exception class for when storage backend is not supported. Passes"""

    def __init__(self, message):
        """Constructor method.

        :param message: The exception message.
        :type message: str

        """
        self.message = message
        super().__init__(self.message)


class MisconfiguredStorageBackend(Exception):
    """Exception class for when storage backend is misconfigured.

    :param message: The exception message.
    :type message: str
    """

    def __init__(self, message):
        """Constructor method."""
        self.message = message
        super().__init__(self.message)


class NoModelException(Exception):
    """Exception class for when model is None

    :param message: The exception message.
    :type message: str
    """

    def __init__(self, message):
        """Constructor method."""
        self.message = message
        super().__init__(self.message)


class CombinersNotDoneException(Exception):
    """ Exception class for when model is None """

    def __init__(self, message):
        """ Constructor method.

        :param message: The exception message.
        :type message: str

        """
        self.message = message
        super().__init__(self.message)


class Control(ControlBase):
    """Controller, implementing the overall global training, validation and inference logic.

    :param statestore: A StateStorage instance.
    :type statestore: class: `fedn.network.statestorebase.StateStorageBase`
    """

    def __init__(self, statestore):
        """Constructor method."""

        super().__init__(statestore)
        self.name = "DefaultControl"

    def session(self, config):
        """Execute a new training session. A session consists of one
            or several global rounds. All rounds in the same session
            have the same round_config.

        :param config: The session config.
        :type config: dict

        """

        if self._state == ReducerState.instructing:
            logger.info("Controller already in INSTRUCTING state. A session is in progress.")
            return

        if not self.statestore.get_latest_model():
            logger.warning("No model in model chain, please provide a seed model!")
            return

        self._state = ReducerState.instructing
        config["committed_at"] = datetime.datetime.now()

        self.create_session(config)

        self._state = ReducerState.monitoring

        last_round = int(self.get_latest_round_id())

        for combiner in self.network.get_combiners():
            combiner.set_aggregator(config['aggregator'])

        self.set_session_status(config['session_id'], 'Started')
        # Execute the rounds in this session
        for round in range(1, int(config["rounds"] + 1)):
            # Increment the round number
            if last_round:
                current_round = last_round + round
            else:
                current_round = round

            try:
                _, round_data = self.round(config, str(current_round))
            except TypeError as e:
                logger.error("Failed to execute round: {0}".format(e))

            logger.info("Round completed with status {}".format(round_data["status"]))

            config["model_id"] = self.statestore.get_latest_model()

        # TODO: Report completion of session
        self.set_session_status(config['session_id'], 'Finished')
        self._state = ReducerState.idle

    def round(self, session_config, round_id):
        """ Execute one global round.

        : param session_config: The session config.
        : type session_config: dict
        : param round_id: The round id.
        : type round_id: str

        """

        self.create_round({'round_id': round_id, 'status': "Pending"})

        if len(self.network.get_combiners()) < 1:
            logger.warning("Round cannot start, no combiners connected!", flush=True)
            self.set_round_status(round_id, 'Failed')
            return None, self.statestore.get_round(round_id)

        # Assemble round config for this global round
        round_config = copy.deepcopy(session_config)
        round_config["rounds"] = 1
        round_config["round_id"] = round_id
        round_config["task"] = "training"

        self.set_round_config(round_id, round_config)

        # Get combiners that are able to participate in the round, given round_config
        participating_combiners = self.get_participating_combiners(round_config)

        # Check if the policy to start the round is met
        round_start = self.evaluate_round_start_policy(participating_combiners)

        if round_start:
            logger.info("round start policy met, {} participating combiners.".format(
                len(participating_combiners)))
        else:
            logger.warning("Round start policy not met, skipping round!")
            self.set_round_status(round_id, 'Failed')
            return None, self.statestore.get_round(round_id)

        # Ask participating combiners to coordinate model updates
        _ = self.request_model_updates(participating_combiners)
        # TODO: Check response

        # Wait until participating combiners have produced an updated global model,
        # or round times out.
        def do_if_round_times_out(result):
            logger.warning("Round timed out!")

        @ retry(wait=wait_random(min=1.0, max=2.0),
                stop=stop_after_delay(session_config['round_timeout']),
                retry_error_callback=do_if_round_times_out,
                retry=retry_if_exception_type(CombinersNotDoneException))
        def combiners_done():

            round = self.statestore.get_round(round_id)
            if 'combiners' not in round:
                logger.info("Waiting for combiners to update model...")
                raise CombinersNotDoneException("Combiners have not yet reported.")

            if len(round['combiners']) < len(participating_combiners):
                logger.info("Waiting for combiners to update model...")
                raise CombinersNotDoneException("All combiners have not yet reported.")

            return True

        combiners_done()

        # Due to the distributed nature of the computation, there might be a
        # delay before combiners have reported the round data to the db,
        # so we need some robustness here.
        @ retry(wait=wait_random(min=0.1, max=1.0),
                retry=retry_if_exception_type(KeyError))
        def check_combiners_done_reporting():
            round = self.statestore.get_round(round_id)
            combiners = round['combiners']
            return combiners

        _ = check_combiners_done_reporting()

        round = self.statestore.get_round(round_id)
        round_valid = self.evaluate_round_validity_policy(round)
        if not round_valid:
            logger.error("Round failed. Invalid - evaluate_round_validity_policy: False")
            self.set_round_status(round_id, 'Failed')
            return None, self.statestore.get_round(round_id)

        logger.info("Reducing combiner level models...")
        # Reduce combiner models into a new global model
        round_data = {}
        try:
            round = self.statestore.get_round(round_id)
            model, data = self.reduce(round['combiners'])
            round_data['reduce'] = data
            logger.info("Done reducing models from combiners!")
        except Exception as e:
            logger.error("Failed to reduce models from combiners, reason: {}".format(e))
            self.set_round_status(round_id, 'Failed')
            return None, self.statestore.get_round(round_id)

        # Commit the new global model to the model trail
        if model is not None:
            logger.info("Committing global model to model trail...")
            tic = time.time()
            model_id = uuid.uuid4()
            session_id = (
                session_config["session_id"]
                if "session_id" in session_config
                else None
            )
            self.commit(model_id, model, session_id)
            round_data["time_commit"] = time.time() - tic
            logger.info("Done committing global model to model trail.")
        else:
            logger.error("Failed to commit model to global model trail.")
            self.set_round_status(round_id, 'Failed')
            return None, self.statestore.get_round(round_id)

        self.set_round_status(round_id, 'Success')

        # 4. Trigger participating combiner nodes to execute a validation round for the current model
        validate = session_config["validate"]
        if validate:
            combiner_config = copy.deepcopy(session_config)
            combiner_config["round_id"] = round_id
            combiner_config["model_id"] = self.statestore.get_latest_model()
            combiner_config["task"] = "validation"
            combiner_config["helper_type"] = self.statestore.get_helper()

            validating_combiners = self.get_participating_combiners(
                combiner_config)

            for combiner, combiner_config in validating_combiners:
                try:
                    logger.info("Submitting validation round to combiner {}".format(combiner))
                    combiner.submit(combiner_config)
                except CombinerUnavailableError:
                    self._handle_unavailable_combiner(combiner)
                    pass

        self.set_round_data(round_id, round_data)
        self.set_round_status(round_id, 'Finished')
        return model_id, self.statestore.get_round(round_id)

    def reduce(self, combiners):
        """Combine updated models from Combiner nodes into one global model.

        : param combiners: dict of combiner names(key) and model IDs(value) to reduce
        : type combiners: dict
        """

        meta = {}
        meta["time_fetch_model"] = 0.0
        meta["time_load_model"] = 0.0
        meta["time_aggregate_model"] = 0.0

        i = 1
        model = None

        for combiner in combiners:
            name = combiner['name']
            model_id = combiner['model_id']

            logger.info("Fetching model ({}) from model repository".format(model_id))

            try:
                tic = time.time()
                data = self.model_repository.get_model(model_id)
                meta['time_fetch_model'] += (time.time() - tic)
            except Exception as e:
                logger.error("Failed to fetch model from model repository {}: {}".format(name, e))
                data = None

            if data is not None:
                try:
                    tic = time.time()
                    helper = self.get_helper()
                    model_next = load_model_from_BytesIO(data, helper)
                    meta["time_load_model"] += time.time() - tic
                    tic = time.time()
                    model = helper.increment_average(model, model_next, 1.0, i)
                    meta["time_aggregate_model"] += time.time() - tic
                except Exception:
                    tic = time.time()
                    model = load_model_from_BytesIO(data, helper)
                    meta["time_aggregate_model"] += time.time() - tic
                i = i + 1

            self.model_repository.delete_model(model_id)

        return model, meta

    def infer_instruct(self, config):
        """Main entrypoint for executing the inference compute plan.

        : param config: configuration for the inference round
        """

        # Check/set instucting state
        if self.__state == ReducerState.instructing:
            logger.info("Already set in INSTRUCTING state")
            return
        self.__state = ReducerState.instructing

        # Check for a model chain
        if not self.statestore.latest_model():
            logger.warning("No model in model chain, please set seed model.")

        # Set reducer in monitoring state
        self.__state = ReducerState.monitoring

        # Start inference round
        try:
            self.inference_round(config)
        except TypeError:
            logger.error("Round failed.")

        # Set reducer in idle state
        self.__state = ReducerState.idle

    def inference_round(self, config):
        """Execute an inference round.

        : param config: configuration for the inference round
        """

        # Init meta
        round_data = {}

        # Check for at least one combiner in statestore
        if len(self.network.get_combiners()) < 1:
            logger.warning("No combiners connected!")
            return round_data

        # Setup combiner configuration
        combiner_config = copy.deepcopy(config)
        combiner_config["model_id"] = self.statestore.get_latest_model()
        combiner_config["task"] = "inference"
        combiner_config["helper_type"] = self.statestore.get_framework()

        # Select combiners
        validating_combiners = self.get_participating_combiners(
            combiner_config)

        # Test round start policy
        round_start = self.check_round_start_policy(validating_combiners)
        if round_start:
            logger.info("Round start policy met, participating combiners {}".format(validating_combiners))
        else:
            logger.warning("Round start policy not met, skipping round!")
            return None

        # Synch combiners with latest model and trigger inference
        for combiner, combiner_config in validating_combiners:
            try:
                combiner.submit(combiner_config)
            except CombinerUnavailableError:
                # It is OK if inference fails for a combiner
                self._handle_unavailable_combiner(combiner)
                pass

        return round_data



from fedn.network.combiner.interfaces import CombinerUnavailableError
from fedn.network.loadbalancer.loadbalancerbase import LoadBalancerBase


class LeastPacked(LoadBalancerBase):
    """ Load balancer that selects the combiner with the least number of attached training clients.

    :param network: A handle to the network.
    :type network: class: `fedn.network.api.network.Network`
    """

    def __init__(self, network):
        super().__init__(network)

    def find_combiner(self):
        """
            Find the combiner with the least number of attached clients.

        """
        min_clients = None
        selected_combiner = None

        for combiner in self.network.get_combiners():
            try:
                if combiner.allowing_clients():
                    # Using default default Channel = 1, MODEL_UPDATE_REQUESTS
                    nr_active_clients = len(combiner.list_active_clients())
                    if not min_clients:
                        min_clients = nr_active_clients
                        selected_combiner = combiner
                    elif nr_active_clients < min_clients:
                        min_clients = nr_active_clients
                        selected_combiner = combiner
            except CombinerUnavailableError:
                pass

        return selected_combiner



from abc import ABC, abstractmethod


class ModelStorage(ABC):

    @abstractmethod
    def exist(self, model_id):
        """ Check if model exists in storage

        :param model_id: The model id
        :type model_id: str
        :return: True if model exists, False otherwise
        :rtype: bool
        """
        pass

    @abstractmethod
    def get(self, model_id):
        """ Get model from storage

        :param model_id: The model id
        :type model_id: str
        :return: The model
        :rtype: object
        """
        pass

    @abstractmethod
    def get_model_metadata(self, model_id):
        """ Get model metadata from storage

        :param model_id: The model id
        :type model_id: str
        :return: The model metadata
        :rtype: dict
        """
        pass

    @abstractmethod
    def set_model_metadata(self, model_id, model_metadata):
        """ Set model metadata in storage

        :param model_id: The model id
        :type model_id: str
        :param model_metadata: The model metadata
        :type model_metadata: dict
        :return: True if successful, False otherwise
        :rtype: bool
        """
        pass

    @abstractmethod
    def delete(self, model_id):
        """ Delete model from storage

        :param model_id: The model id
        :type model_id: str
        :return: True if successful, False otherwise
        :rtype: bool
        """
        pass

    @abstractmethod
    def delete_all(self):
        """ Delete all models from storage

        :return: True if successful, False otherwise
        :rtype: bool
        """
        pass



# Unittest for Flask API endpoints
#
# Run with:
# python -m unittest fedn.tests.network.api.tests
#
# or
#
# python3 -m unittest fedn.tests.network.api.tests
#
# or
#
# python3 -m unittest fedn.tests.network.api.tests.NetworkAPITests
#
# or
#
# python -m unittest fedn.tests.network.api.tests.NetworkAPITests
#
# or
#
# python -m unittest fedn.tests.network.api.tests.NetworkAPITests.test_get_model_trail
#
# or
#
# python3 -m unittest fedn.tests.network.api.tests.NetworkAPITests.test_get_model_trail
#

import io
import time
import unittest
from unittest.mock import MagicMock, patch

import fedn


class NetworkAPITests(unittest.TestCase):
    """ Unittests for the Network API. """
    @patch('fedn.network.statestore.mongostatestore.MongoStateStore', autospec=True)
    @patch('fedn.network.controller.controlbase.ControlBase', autospec=True)
    def setUp(self, mock_mongo, mock_control):
        # start Flask server in testing mode
        import fedn.network.api.server
        self.app = fedn.network.api.server.app.test_client()

    def test_get_model_trail(self):
        """ Test get_model_trail endpoint. """
        # Mock api.get_model_trail
        model_id = "test"
        time_stamp = time.time()
        return_value = {model_id: time_stamp}
        fedn.network.api.server.api.get_model_trail = MagicMock(return_value=return_value)
        # Make request
        response = self.app.get('/get_model_trail')
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.get_model_trail was called
        fedn.network.api.server.api.get_model_trail.assert_called_once_with()

    def test_get_latest_model(self):
        """ Test get_latest_model endpoint. """
        # Mock api.get_latest_model
        model_id = "test"
        time_stamp = time.time()
        return_value = {model_id: time_stamp}
        fedn.network.api.server.api.get_latest_model = MagicMock(return_value=return_value)
        # Make request
        response = self.app.get('/get_latest_model')
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.get_latest_model was called
        fedn.network.api.server.api.get_latest_model.assert_called_once_with()

    def test_get_initial_model(self):
        """ Test get_initial_model endpoint. """
        # Mock api.get_initial_model
        model_id = "test"
        time_stamp = time.time()
        return_value = {model_id: time_stamp}
        fedn.network.api.server.api.get_initial_model = MagicMock(return_value=return_value)
        # Make request
        response = self.app.get('/get_initial_model')
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.get_initial_model was called
        fedn.network.api.server.api.get_initial_model.assert_called_once_with()

    def test_set_initial_model(self):
        """ Test set_initial_model endpoint. """
        # Mock api.set_initial_model
        success = True
        message = "test"
        return_value = {'success': success, 'message': message}
        fedn.network.api.server.api.set_initial_model = MagicMock(return_value=return_value)
        # Create test file
        request_file = (io.BytesIO(b"abcdef"), 'test.txt')
        # Make request
        response = self.app.post('/set_initial_model', data={"file": request_file})
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.set_initial_model was called
        fedn.network.api.server.api.set_initial_model.assert_called_once()

    def test_list_clients(self):
        """ Test list_clients endpoint. """
        # Mock api.get_all_clients
        return_value = {"test": "test"}
        fedn.network.api.server.api.get_all_clients = MagicMock(return_value=return_value)
        # Make request
        response = self.app.get('/list_clients')
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.get_all_clients was called
        fedn.network.api.server.api.get_all_clients.assert_called_once_with()

    def test_get_active_clients(self):
        """ Test get_active_clients endpoint. """
        # Mock api.get_active_clients
        return_value = {"test": "test"}
        fedn.network.api.server.api.get_active_clients = MagicMock(return_value=return_value)
        # Make request
        response = self.app.get('/get_active_clients?combiner=test')
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.get_active_clients was called
        fedn.network.api.server.api.get_active_clients.assert_called_once_with("test")

    def test_add_client(self):
        """ Test add_client endpoint. """
        # Mock api.add_client
        return_value = {"test": "test"}
        fedn.network.api.server.api.add_client = MagicMock(return_value=return_value)
        # Make request
        response = self.app.post('/add_client', json={
            'preferred_combiner': 'test',
        })
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.add_client was called
        fedn.network.api.server.api.add_client.assert_called_once_with(
            preferred_combiner="test",
            remote_addr='127.0.0.1'
        )

    def test_list_combiners(self):
        """ Test list_combiners endpoint. """
        # Mock api.get_all_combiners
        return_value = {"test": "test"}
        fedn.network.api.server.api.get_all_combiners = MagicMock(return_value=return_value)
        # Make request
        response = self.app.get('/list_combiners')
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.get_all_combiners was called
        fedn.network.api.server.api.get_all_combiners.assert_called_once_with()

    def test_list_compute_packages(self):
        """ Test list_compute_packages endpoint. """
        # Mock api.list_compute_packages
        return_value = {"test": "test"}
        fedn.network.api.server.api.list_compute_packages = MagicMock(return_value=return_value)
        # Make request
        response = self.app.get('/list_combiners')
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.list_compute_packages was called
        fedn.network.api.server.api.list_compute_packages.assert_called_once_with()

    def test_list_rounds(self):
        """ Test list_rounds endpoint. """
        # Mock api.get_all_rounds
        return_value = {"test": "test"}
        fedn.network.api.server.api.get_all_rounds = MagicMock(return_value=return_value)
        # Make request
        response = self.app.get('/list_rounds')
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.get_all_rounds was called
        fedn.network.api.server.api.get_all_rounds.assert_called_once_with()

    def test_get_round(self):
        """ Test get_round endpoint. """
        # Mock api.get_round
        return_value = {"test": "test"}
        fedn.network.api.server.api.get_round = MagicMock(return_value=return_value)
        # Make request
        response = self.app.get('/get_round?round_id=test')
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.get_round was called
        fedn.network.api.server.api.get_round.assert_called_once_with("test")

    def test_get_combiner(self):
        """ Test get_combiner endpoint. """
        # Mock api.get_combiner
        return_value = {"test": "test"}
        fedn.network.api.server.api.get_combiner = MagicMock(return_value=return_value)
        # Make request
        response = self.app.get('/get_combiner?combiner=test')
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.get_combiner was called
        fedn.network.api.server.api.get_combiner.assert_called_once_with("test")

    def test_add_combiner(self):
        """ Test add_combiner endpoint. """
        # Mock api.add_combiner
        success = True
        message = "test"
        return_value = {'success': success, 'message': message}
        fedn.network.api.server.api.add_combiner = MagicMock(return_value=return_value)
        # Make request
        response = self.app.post('/add_combiner', json={
            'combiner_id': 'test',
            'address': '1234',
            'port': '1234',
            'secure_grpc': 'True',
            'fqdn': 'test',
        })
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.add_combiner was called
        fedn.network.api.server.api.add_combiner.assert_called_once_with(
            combiner_id='test',
            remote_addr='127.0.0.1',
            address='1234',
            port='1234',
            secure_grpc='True',
            fqdn='test',
        )

    def test_get_events(self):
        """ Test get_events endpoint. """
        # Mock api.get_events
        return_value = {"test": "test"}
        fedn.network.api.server.api.get_events = MagicMock(return_value=return_value)
        # Make request
        response = self.app.get('/get_events')
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.get_events was called
        fedn.network.api.server.api.get_events.assert_called_once()

    def test_get_status(self):
        """ Test get_status endpoint. """
        # Mock api.get_status
        return_value = {"test": "test"}
        fedn.network.api.server.api.get_controller_status = MagicMock(return_value=return_value)
        # Make request
        response = self.app.get('/get_controller_status')
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.get_status was called
        fedn.network.api.server.api.get_controller_status.assert_called_once()

    def test_start_session(self):
        """ Test start_session endpoint. """
        # Mock api.start_session
        success = True
        message = "test"
        return_value = {'success': success, 'message': message}
        fedn.network.api.server.api.start_session = MagicMock(return_value=return_value)
        # Make request with only session_id
        json = {'session_id': 'test',
                'round_timeout': float(60),
                'rounds': 1,
                'round_buffer_size': -1,
                }
        response = self.app.post('/start_session', json=json)
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.start_session was called
        fedn.network.api.server.api.start_session.assert_called_once_with(
            session_id='test',
            round_timeout=float(60),
            rounds=1,
            round_buffer_size=-1,
        )

    def test_list_sessions(self):
        """ Test list_sessions endpoint. """
        # Mock api.list_sessions
        return_value = {"test": "test"}
        fedn.network.api.server.api.get_all_sessions = MagicMock(return_value=return_value)
        # Make request
        response = self.app.get('/list_sessions')
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.list_sessions was called
        fedn.network.api.server.api.get_all_sessions.assert_called_once()

    def test_list_models(self):
        """ Test list_models endpoint. """
        # Mock api.list_models
        return_value = {"test": "test"}
        fedn.network.api.server.api.get_models = MagicMock(return_value=return_value)
        # Make request
        response = self.app.get('/list_models')
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.list_models was called
        fedn.network.api.server.api.get_models.assert_called_once()

    def test_get_package(self):
        """ Test get_package endpoint. """
        # Mock api.get_package
        return_value = {"test": "test"}
        fedn.network.api.server.api.get_compute_package = MagicMock(return_value=return_value)
        # Make request
        response = self.app.get('/get_package')
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.get_package was called
        fedn.network.api.server.api.get_compute_package.assert_called_once_with()

    def test_get_controller_status(self):
        """ Test get_controller_status endpoint. """
        # Mock api.get_controller_status
        return_value = {"test": "test"}
        fedn.network.api.server.api.get_controller_status = MagicMock(return_value=return_value)
        # Make request
        response = self.app.get('/get_controller_status')
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.get_controller_status was called
        fedn.network.api.server.api.get_controller_status.assert_called_once_with()

    def test_get_client_config(self):
        """ Test get_client_config endpoint. """
        # Mock api.get_client_config
        return_value = {"test": "test"}
        fedn.network.api.server.api.get_client_config = MagicMock(return_value=return_value)
        # Make request
        response = self.app.get('/get_client_config')
        # Assert response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json, return_value)
        # Assert api.get_client_config was called
        fedn.network.api.server.api.get_client_config.assert_called_once_with(True)


if __name__ == '__main__':
    unittest.main()



import subprocess

from fedn.common.log_config import logger


def run_process(args, cwd):
    """ Run a process and log the output.

    :param args: The arguments to the process.
    :type args: list
    :param cwd: The current working directory.
    :type cwd: str
    :return:
    """
    status = subprocess.Popen(
        args, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    # print(status)
    def check_io():
        """ Check stdout/stderr of the child process.

        :return:
        """
        while True:
            output = status.stdout.readline().decode()
            if output:
                logger.info(output)
            else:
                break

    # keep checking stdout/stderr until the child exits
    while status.poll() is None:
        check_io()



import base64
import io
import json
import os
import queue
import re
import socket
import sys
import tempfile
import threading
import time
import uuid
from datetime import datetime
from distutils.dir_util import copy_tree
from io import BytesIO

import grpc
from cryptography.hazmat.primitives.serialization import Encoding
from google.protobuf.json_format import MessageToJson
from OpenSSL import SSL

import fedn.network.grpc.fedn_pb2 as fedn
import fedn.network.grpc.fedn_pb2_grpc as rpc
from fedn.common.log_config import (logger, set_log_level_from_string,
                                    set_log_stream)
from fedn.network.clients.connect import ConnectorClient, Status
from fedn.network.clients.package import PackageRuntime
from fedn.network.clients.state import ClientState, ClientStateToString
from fedn.network.combiner.modelservice import upload_request_generator
from fedn.utils.dispatcher import Dispatcher
from fedn.utils.helpers.helpers import get_helper

CHUNK_SIZE = 1024 * 1024
VALID_NAME_REGEX = '^[a-zA-Z0-9_-]*$'


class GrpcAuth(grpc.AuthMetadataPlugin):
    def __init__(self, key):
        self._key = key

    def __call__(self, context, callback):
        callback((('authorization', f'Token {self._key}'),), None)


class Client:
    """FEDn Client. Service running on client/datanodes in a federation,
    recieving and handling model update and model validation requests.

    :param config: A configuration dictionary containing connection information for the discovery service (controller)
        and settings governing e.g. client-combiner assignment behavior.
    :type config: dict
    """

    def __init__(self, config):
        """Initialize the client."""
        self.state = None
        self.error_state = False
        self._attached = False
        self._missed_heartbeat = 0
        self.config = config

        set_log_level_from_string(config.get('verbosity', "INFO"))
        set_log_stream(config.get('logfile', None))

        self.connector = ConnectorClient(host=config['discover_host'],
                                         port=config['discover_port'],
                                         token=config['token'],
                                         name=config['name'],
                                         remote_package=config['remote_compute_context'],
                                         force_ssl=config['force_ssl'],
                                         verify=config['verify'],
                                         combiner=config['preferred_combiner'],
                                         id=config['client_id'])

        # Validate client name
        match = re.search(VALID_NAME_REGEX, config['name'])
        if not match:
            raise ValueError('Unallowed character in client name. Allowed characters: a-z, A-Z, 0-9, _, -.')

        self.name = config['name']
        dirname = time.strftime("%Y%m%d-%H%M%S")
        self.run_path = os.path.join(os.getcwd(), dirname)
        os.mkdir(self.run_path)

        self.started_at = datetime.now()
        self.logs = []

        self.inbox = queue.Queue()

        # Attach to the FEDn network (get combiner)
        client_config = self._attach()

        self._initialize_dispatcher(config)

        self._initialize_helper(client_config)
        if not self.helper:
            logger.warning("Failed to retrieve helper class settings: {}".format(
                client_config))

        self._subscribe_to_combiner(config)

        self.state = ClientState.idle

    def _assign(self):
        """Contacts the controller and asks for combiner assignment.

        :return: A configuration dictionary containing connection information for combiner.
        :rtype: dict
        """

        logger.info("Initiating assignment request.")
        while True:
            status, response = self.connector.assign()
            if status == Status.TryAgain:
                logger.info(response)
                time.sleep(5)
                continue
            if status == Status.Assigned:
                client_config = response
                break
            if status == Status.UnAuthorized:
                logger.critical(response)
                sys.exit("Exiting: Unauthorized")
            if status == Status.UnMatchedConfig:
                logger.critical(response)
                sys.exit("Exiting: UnMatchedConfig")
            time.sleep(5)

        logger.info("Assignment successfully received.")
        logger.info("Received combiner configuration: {}".format(client_config))
        return client_config

    def _add_grpc_metadata(self, key, value):
        """Add metadata for gRPC calls.

        :param key: The key of the metadata.
        :type key: str
        :param value: The value of the metadata.
        :type value: str
        """
        # Check if metadata exists and add if not
        if not hasattr(self, 'metadata'):
            self.metadata = ()

        # Check if metadata key already exists and replace value if so
        for i, (k, v) in enumerate(self.metadata):
            if k == key:
                # Replace value
                self.metadata = self.metadata[:i] + ((key, value),) + self.metadata[i + 1:]
                return

        # Set metadata using tuple concatenation
        self.metadata += ((key, value),)

    def _get_ssl_certificate(self, domain, port=443):
        context = SSL.Context(SSL.SSLv23_METHOD)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((domain, port))
        ssl_sock = SSL.Connection(context, sock)
        ssl_sock.set_tlsext_host_name(domain.encode())
        ssl_sock.set_connect_state()
        ssl_sock.do_handshake()
        cert = ssl_sock.get_peer_certificate()
        ssl_sock.close()
        sock.close()
        cert = cert.to_cryptography().public_bytes(Encoding.PEM).decode()
        return cert

    def _connect(self, client_config):
        """Connect to assigned combiner.

        :param client_config: A configuration dictionary containing connection information for
        the combiner.
        :type client_config: dict
        """

        # TODO use the client_config['certificate'] for setting up secure comms'
        host = client_config['host']
        # Add host to gRPC metadata
        self._add_grpc_metadata('grpc-server', host)
        logger.info("Client using metadata: {}.".format(self.metadata))
        port = client_config['port']
        secure = False
        if client_config['fqdn'] is not None:
            host = client_config['fqdn']
            # assuming https if fqdn is used
            port = 443
        logger.info(f"Initiating connection to combiner host at: {host}:{port}")

        if client_config['certificate']:
            logger.info("Utilizing CA certificate for GRPC channel authentication.")
            secure = True
            cert = base64.b64decode(
                client_config['certificate'])  # .decode('utf-8')
            credentials = grpc.ssl_channel_credentials(root_certificates=cert)
            channel = grpc.secure_channel("{}:{}".format(host, str(port)), credentials)
        elif os.getenv("FEDN_GRPC_ROOT_CERT_PATH"):
            secure = True
            logger.info("Using root certificate from environment variable for GRPC channel.")
            with open(os.environ["FEDN_GRPC_ROOT_CERT_PATH"], 'rb') as f:
                credentials = grpc.ssl_channel_credentials(f.read())
            channel = grpc.secure_channel("{}:{}".format(host, str(port)), credentials)
        elif self.config['secure']:
            secure = True
            logger.info("Using CA certificate for GRPC channel.")
            cert = self._get_ssl_certificate(host, port=port)

            credentials = grpc.ssl_channel_credentials(cert.encode('utf-8'))
            if self.config['token']:
                token = self.config['token']
                auth_creds = grpc.metadata_call_credentials(GrpcAuth(token))
                channel = grpc.secure_channel("{}:{}".format(host, str(port)), grpc.composite_channel_credentials(credentials, auth_creds))
            else:
                channel = grpc.secure_channel("{}:{}".format(host, str(port)), credentials)
        else:
            logger.info("Using insecure GRPC channel.")
            if port == 443:
                port = 80
            channel = grpc.insecure_channel("{}:{}".format(
                host,
                str(port)))

        self.channel = channel

        self.connectorStub = rpc.ConnectorStub(channel)
        self.combinerStub = rpc.CombinerStub(channel)
        self.modelStub = rpc.ModelServiceStub(channel)

        logger.info("Successfully established {} connection to {}:{}".format("secure" if secure else "insecure",
                                                                             host,
                                                                             port))

        logger.info("Using {} compute package.".format(
            client_config["package"]))

    def _disconnect(self):
        """Disconnect from the combiner."""
        self.channel.close()

    def detach(self):
        """Detach from the FEDn network (disconnect from combiner)"""
        # Setting _attached to False will make all processing threads return
        if not self._attached:
            logger.info("Client is not attached.")

        self._attached = False
        # Close gRPC connection to combiner
        self._disconnect()

    def _attach(self):
        """Attach to the FEDn network (connect to combiner)"""
        # Ask controller for a combiner and connect to that combiner.
        if self._attached:
            logger.info("Client is already attached. ")
            return None

        client_config = self._assign()
        self._connect(client_config)

        if client_config:
            self._attached = True
        return client_config

    def _initialize_helper(self, client_config):
        """Initialize the helper class for the client.

        :param client_config: A configuration dictionary containing connection information for
        | the discovery service (controller) and settings governing e.g.
        | client-combiner assignment behavior.
        :type client_config: dict
        :return:
        """

        if 'helper_type' in client_config.keys():
            self.helper = get_helper(client_config['helper_type'])

    def _subscribe_to_combiner(self, config):
        """Listen to combiner message stream and start all processing threads.

        :param config: A configuration dictionary containing connection information for
        | the discovery service (controller) and settings governing e.g.
        | client-combiner assignment behavior.
        """

        # Start sending heartbeats to the combiner.
        threading.Thread(target=self._send_heartbeat, kwargs={
            'update_frequency': config['heartbeat_interval']}, daemon=True).start()

        # Start listening for combiner training and validation messages
        threading.Thread(
            target=self._listen_to_task_stream, daemon=True).start()
        self._attached = True

        # Start processing the client message inbox
        threading.Thread(target=self.process_request, daemon=True).start()

    def _initialize_dispatcher(self, config):
        """ Initialize the dispatcher for the client.

        :param config: A configuration dictionary containing connection information for
        | the discovery service (controller) and settings governing e.g.
        | client-combiner assignment behavior.
        :type config: dict
        :return:
        """
        if config['remote_compute_context']:
            pr = PackageRuntime(os.getcwd(), os.getcwd())

            retval = None
            tries = 10

            while tries > 0:
                retval = pr.download(
                    host=config['discover_host'],
                    port=config['discover_port'],
                    token=config['token'],
                    force_ssl=config['force_ssl'],
                    secure=config['secure']
                )
                if retval:
                    break
                time.sleep(60)
                logger.warning("Compute package not available. Retrying in 60 seconds. {} attempts remaining.".format(tries))
                tries -= 1

            if retval:
                if 'checksum' not in config:
                    logger.warning("Bypassing validation of package checksum. Ensure the package source is trusted.")
                else:
                    checks_out = pr.validate(config['checksum'])
                    if not checks_out:
                        logger.critical("Validation of local package failed. Client terminating.")
                        self.error_state = True
                        return

            if retval:
                pr.unpack()

            self.dispatcher = pr.dispatcher(self.run_path)
            try:
                logger.info("Initiating Dispatcher with entrypoint set to: startup")
                self.dispatcher.run_cmd("startup")
            except KeyError:
                pass
            except Exception as e:
                logger.error(f"Caught exception: {type(e).__name__}")
        else:
            # TODO: Deprecate
            dispatch_config = {'entry_points':
                               {'predict': {'command': 'python3 predict.py'},
                                'train': {'command': 'python3 train.py'},
                                'validate': {'command': 'python3 validate.py'}}}
            from_path = os.path.join(os.getcwd(), 'client')

            copy_tree(from_path, self.run_path)
            self.dispatcher = Dispatcher(dispatch_config, self.run_path)

    def get_model_from_combiner(self, id, timeout=20):
        """Fetch a model from the assigned combiner.
        Downloads the model update object via a gRPC streaming channel.

        :param id: The id of the model update object.
        :type id: str
        :return: The model update object.
        :rtype: BytesIO
        """
        data = BytesIO()
        time_start = time.time()
        request = fedn.ModelRequest(id=id)
        request.sender.name = self.name
        request.sender.role = fedn.WORKER

        for part in self.modelStub.Download(request, metadata=self.metadata):

            if part.status == fedn.ModelStatus.IN_PROGRESS:
                data.write(part.data)

            if part.status == fedn.ModelStatus.OK:
                return data

            if part.status == fedn.ModelStatus.FAILED:
                return None

            if part.status == fedn.ModelStatus.UNKNOWN:
                if time.time() - time_start >= timeout:
                    return None
                continue

        return data

    def send_model_to_combiner(self, model, id):
        """Send a model update to the assigned combiner.
        Uploads the model updated object via a gRPC streaming channel, Upload.

        :param model: The model update object.
        :type model: BytesIO
        :param id: The id of the model update object.
        :type id: str
        :return: The model update object.
        :rtype: BytesIO
        """
        if not isinstance(model, BytesIO):
            bt = BytesIO()

            for d in model.stream(32 * 1024):
                bt.write(d)
        else:
            bt = model

        bt.seek(0, 0)

        result = self.modelStub.Upload(upload_request_generator(bt, id), metadata=self.metadata)

        return result

    def _listen_to_task_stream(self):
        """Subscribe to the model update request stream.

        :return: None
        :rtype: None
        """

        r = fedn.ClientAvailableMessage()
        r.sender.name = self.name
        r.sender.role = fedn.WORKER
        # Add client to metadata
        self._add_grpc_metadata('client', self.name)

        while self._attached:
            try:
                for request in self.combinerStub.TaskStream(r, metadata=self.metadata):
                    if request:
                        logger.debug("Received model update request from combiner: {}.".format(request))
                    if request.sender.role == fedn.COMBINER:
                        # Process training request
                        self._send_status("Received model update request.", log_level=fedn.Status.AUDIT,
                                          type=fedn.StatusType.MODEL_UPDATE_REQUEST, request=request)
                        logger.info("Received model update request of type {} for model_id {}".format(request.type, request.model_id))

                        if request.type == fedn.StatusType.MODEL_UPDATE and self.config['trainer']:
                            self.inbox.put(('train', request))
                        elif request.type == fedn.StatusType.MODEL_VALIDATION and self.config['validator']:
                            self.inbox.put(('validate', request))
                        else:
                            logger.error("Unknown request type: {}".format(request.type))

            except grpc.RpcError as e:
                # Handle gRPC errors
                status_code = e.code()
                if status_code == grpc.StatusCode.UNAVAILABLE:
                    logger.warning("GRPC server unavailable during model update request stream. Retrying.")
                    # Retry after a delay
                    time.sleep(5)
                else:
                    # Log the error and continue
                    logger.error(f"An error occurred during model update request stream: {e}")

            except Exception as ex:
                # Handle other exceptions
                logger.error(f"An error occurred during model update request stream: {ex}")

        # Detach if not attached
        if not self._attached:
            return

    def _process_training_request(self, model_id):
        """Process a training (model update) request.

        :param model_id: The model id of the model to be updated.
        :type model_id: str
        :return: The model id of the updated model, or None if the update failed. And a dict with metadata.
        :rtype: tuple
        """

        self._send_status(
            "\t Starting processing of training request for model_id {}".format(model_id))
        self.state = ClientState.training

        try:
            meta = {}
            tic = time.time()
            mdl = self.get_model_from_combiner(str(model_id))
            if mdl is None:
                logger.error("Could not retrieve model from combiner. Aborting training request.")
                return None, None
            meta['fetch_model'] = time.time() - tic

            inpath = self.helper.get_tmp_path()
            with open(inpath, 'wb') as fh:
                fh.write(mdl.getbuffer())

            outpath = self.helper.get_tmp_path()
            tic = time.time()
            # TODO: Check return status, fail gracefully

            self.dispatcher.run_cmd("train {} {}".format(inpath, outpath))

            meta['exec_training'] = time.time() - tic

            tic = time.time()
            out_model = None

            with open(outpath, "rb") as fr:
                out_model = io.BytesIO(fr.read())

            # Stream model update to combiner server
            updated_model_id = uuid.uuid4()
            self.send_model_to_combiner(out_model, str(updated_model_id))
            meta['upload_model'] = time.time() - tic

            # Read the metadata file
            with open(outpath+'-metadata', 'r') as fh:
                training_metadata = json.loads(fh.read())
            meta['training_metadata'] = training_metadata

            os.unlink(inpath)
            os.unlink(outpath)
            os.unlink(outpath+'-metadata')

        except Exception as e:
            logger.error("Could not process training request due to error: {}".format(e))
            updated_model_id = None
            meta = {'status': 'failed', 'error': str(e)}

        self.state = ClientState.idle

        return updated_model_id, meta

    def _process_validation_request(self, model_id, is_inference):
        """Process a validation request.

        :param model_id: The model id of the model to be validated.
        :type model_id: str
        :param is_inference: True if the validation is an inference request, False if it is a validation request.
        :type is_inference: bool
        :return: The validation metrics, or None if validation failed.
        :rtype: dict
        """
        # Figure out cmd
        if is_inference:
            cmd = 'infer'
        else:
            cmd = 'validate'

        self._send_status(
            f"Processing {cmd} request for model_id {model_id}")
        self.state = ClientState.validating
        try:
            model = self.get_model_from_combiner(str(model_id))
            if model is None:
                logger.error("Could not retrieve model from combiner. Aborting validation request.")
                return None
            inpath = self.helper.get_tmp_path()

            with open(inpath, "wb") as fh:
                fh.write(model.getbuffer())

            _, outpath = tempfile.mkstemp()
            self.dispatcher.run_cmd(f"{cmd} {inpath} {outpath}")

            with open(outpath, "r") as fh:
                validation = json.loads(fh.read())

            os.unlink(inpath)
            os.unlink(outpath)

        except Exception as e:
            logger.warning("Validation failed with exception {}".format(e))
            self.state = ClientState.idle
            return None

        self.state = ClientState.idle
        return validation

    def process_request(self):
        """Process training and validation tasks. """
        while True:

            if not self._attached:
                return

            try:
                (task_type, request) = self.inbox.get(timeout=1.0)
                if task_type == 'train':

                    tic = time.time()
                    self.state = ClientState.training
                    model_id, meta = self._process_training_request(
                        request.model_id)
                    processing_time = time.time()-tic
                    meta['processing_time'] = processing_time
                    meta['config'] = request.data

                    if model_id is not None:
                        # Send model update to combiner
                        update = fedn.ModelUpdate()
                        update.sender.name = self.name
                        update.sender.role = fedn.WORKER
                        update.receiver.name = request.sender.name
                        update.receiver.role = request.sender.role
                        update.model_id = request.model_id
                        update.model_update_id = str(model_id)
                        update.timestamp = str(datetime.now())
                        update.correlation_id = request.correlation_id
                        update.meta = json.dumps(meta)
                        # TODO: Check responses
                        _ = self.combinerStub.SendModelUpdate(update, metadata=self.metadata)
                        self._send_status("Model update completed.", log_level=fedn.Status.AUDIT,
                                          type=fedn.StatusType.MODEL_UPDATE, request=update)

                    else:
                        self._send_status("Client {} failed to complete model update.",
                                          log_level=fedn.Status.WARNING,
                                          request=request)
                    self.state = ClientState.idle
                    self.inbox.task_done()

                elif task_type == 'validate':
                    self.state = ClientState.validating
                    metrics = self._process_validation_request(
                        request.model_id, False)

                    if metrics is not None:
                        # Send validation
                        validation = fedn.ModelValidation()
                        validation.sender.name = self.name
                        validation.sender.role = fedn.WORKER
                        validation.receiver.name = request.sender.name
                        validation.receiver.role = request.sender.role
                        validation.model_id = str(request.model_id)
                        validation.data = json.dumps(metrics)
                        validation.timestamp.GetCurrentTime()
                        validation.correlation_id = request.correlation_id

                        _ = self.combinerStub.SendModelValidation(
                            validation, metadata=self.metadata)

                        status_type = fedn.StatusType.MODEL_VALIDATION

                        self._send_status("Model validation completed.", log_level=fedn.Status.AUDIT,
                                          type=status_type, request=validation)
                    else:
                        self._send_status("Client {} failed to complete model validation.".format(self.name),
                                          log_level=fedn.Status.WARNING, request=request)

                    self.state = ClientState.idle
                    self.inbox.task_done()
            except queue.Empty:
                pass

    def _handle_combiner_failure(self):
        """ Register failed combiner connection."""
        self._missed_heartbeat += 1
        if self._missed_heartbeat > self.config['reconnect_after_missed_heartbeat']:
            self.detach()()

    def _send_heartbeat(self, update_frequency=2.0):
        """Send a heartbeat to the combiner.

        :param update_frequency: The frequency of the heartbeat in seconds.
        :type update_frequency: float
        :return: None if the client is detached.
        :rtype: None
        """
        while True:
            heartbeat = fedn.Heartbeat(sender=fedn.Client(
                name=self.name, role=fedn.WORKER))
            try:
                self.connectorStub.SendHeartbeat(heartbeat, metadata=self.metadata)
                self._missed_heartbeat = 0
            except grpc.RpcError as e:
                status_code = e.code()
                logger.warning("Client heartbeat: GRPC error, {}. Retrying.".format(
                    status_code.name))
                logger.debug(e)
                self._handle_combiner_failure()

            time.sleep(update_frequency)
            if not self._attached:
                return

    def _send_status(self, msg, log_level=fedn.Status.INFO, type=None, request=None):
        """Send status message.

        :param msg: The message to send.
        :type msg: str
        :param log_level: The log level of the message.
        :type log_level: fedn.Status.INFO, fedn.Status.WARNING, fedn.Status.ERROR
        :param type: The type of the message.
        :type type: str
        :param request: The request message.
        :type request: fedn.Request
        """
        status = fedn.Status()
        status.timestamp.GetCurrentTime()
        status.sender.name = self.name
        status.sender.role = fedn.WORKER
        status.log_level = log_level
        status.status = str(msg)
        if type is not None:
            status.type = type

        if request is not None:
            status.data = MessageToJson(request)

        self.logs.append(
            "{} {} LOG LEVEL {} MESSAGE {}".format(str(datetime.now()), status.sender.name, status.log_level,
                                                   status.status))
        _ = self.connectorStub.SendStatus(status, metadata=self.metadata)

    def run(self):
        """ Run the client. """
        try:
            cnt = 0
            old_state = self.state
            while True:
                time.sleep(1)
                if cnt == 0:
                    logger.info("Client is active, waiting for model update requests.")
                    cnt = 1
                if self.state != old_state:
                    logger.info("Client in {} state.".format(ClientStateToString(self.state)))
                if not self._attached:
                    logger.info("Detached from combiner.")
                    # TODO: Implement a check/condition to ulitmately close down if too many reattachment attepts have failed. s
                    self._attach()
                    self._subscribe_to_combiner(self.config)
                if self.error_state:
                    return
        except KeyboardInterrupt:
            logger.info("Shutting down.")



from fedn.common.log_config import logger
from fedn.network.combiner.aggregators.aggregatorbase import AggregatorBase


class Aggregator(AggregatorBase):
    """ Local SGD / Federated Averaging (FedAvg) aggregator. Computes a weighted mean
        of parameter updates.

    :param id: A reference to id of :class: `fedn.network.combiner.Combiner`
    :type id: str
    :param storage: Model repository for :class: `fedn.network.combiner.Combiner`
    :type storage: class: `fedn.common.storage.s3.s3repo.S3ModelRepository`
    :param server: A handle to the Combiner class :class: `fedn.network.combiner.Combiner`
    :type server: class: `fedn.network.combiner.Combiner`
    :param modelservice: A handle to the model service :class: `fedn.network.combiner.modelservice.ModelService`
    :type modelservice: class: `fedn.network.combiner.modelservice.ModelService`
    :param control: A handle to the :class: `fedn.network.combiner.roundhandler.RoundHandler`
    :type control: class: `fedn.network.combiner.roundhandler.RoundHandler`

    """

    def __init__(self, storage, server, modelservice, round_handler):
        """Constructor method"""

        super().__init__(storage, server, modelservice, round_handler)

        self.name = "fedavg"

    def combine_models(self, helper=None, delete_models=True):
        """Aggregate all model updates in the queue by computing an incremental
        weighted average of model parameters.

        :param helper: An instance of :class: `fedn.utils.helpers.helpers.HelperBase`, ML framework specific helper, defaults to None
        :type helper: class: `fedn.utils.helpers.helpers.HelperBase`, optional
        :param time_window: The time window for model aggregation, defaults to 180
        :type time_window: int, optional
        :param max_nr_models: The maximum number of updates aggregated, defaults to 100
        :type max_nr_models: int, optional
        :param delete_models: Delete models from storage after aggregation, defaults to True
        :type delete_models: bool, optional
        :return: The global model and metadata
        :rtype: tuple
        """

        data = {}
        data['time_model_load'] = 0.0
        data['time_model_aggregation'] = 0.0

        model = None
        nr_aggregated_models = 0
        total_examples = 0

        logger.info(
            "AGGREGATOR({}): Aggregating model updates... ".format(self.name))

        while not self.model_updates.empty():
            try:
                # Get next model from queue
                model_update = self.next_model_update()

                # Load model parameters and metadata
                model_next, metadata = self.load_model_update(model_update, helper)

                logger.info(
                    "AGGREGATOR({}): Processing model update {}, metadata: {}  ".format(self.name, model_update.model_update_id, metadata))

                # Increment total number of examples
                total_examples += metadata['num_examples']

                if nr_aggregated_models == 0:
                    model = model_next
                else:
                    model = helper.increment_average(
                        model, model_next, metadata['num_examples'], total_examples)

                nr_aggregated_models += 1
                # Delete model from storage
                if delete_models:
                    self.modelservice.temp_model_storage.delete(model_update.model_update_id)
                    logger.info(
                        "AGGREGATOR({}): Deleted model update {} from storage.".format(self.name, model_update.model_update_id))
                self.model_updates.task_done()
            except Exception as e:
                logger.error(
                    "AGGREGATOR({}): Error encoutered while processing model update {}, skipping this update.".format(self.name, e))
                self.model_updates.task_done()

        data['nr_aggregated_models'] = nr_aggregated_models

        logger.info("AGGREGATOR({}): Aggregation completed, aggregated {} models.".format(self.name, nr_aggregated_models))
        return model, data



import os
import tempfile
from io import BytesIO

import fedn.network.grpc.fedn_pb2 as fedn
import fedn.network.grpc.fedn_pb2_grpc as rpc
from fedn.common.log_config import logger
from fedn.network.storage.models.tempmodelstorage import TempModelStorage

CHUNK_SIZE = 1024 * 1024


def upload_request_generator(mdl, id):
    """Generator function for model upload requests.

    :param mdl: The model update object.
    :type mdl: BytesIO
    :return: A model update request.
    :rtype: fedn.ModelRequest
    """
    while True:
        b = mdl.read(CHUNK_SIZE)
        if b:
            result = fedn.ModelRequest(
                data=b, id=id, status=fedn.ModelStatus.IN_PROGRESS)
        else:
            result = fedn.ModelRequest(
                id=id, data=None, status=fedn.ModelStatus.OK)
        yield result
        if not b:
            break


def model_as_bytesIO(model):
    if not isinstance(model, BytesIO):
        bt = BytesIO()

        written_total = 0
        for d in model.stream(32 * 1024):
            written = bt.write(d)
            written_total += written
    else:
        bt = model

    bt.seek(0, 0)
    return bt


def get_tmp_path():
    """ Return a temporary output path compatible with save_model, load_model. """
    fd, path = tempfile.mkstemp()
    os.close(fd)
    return path


def load_model_from_BytesIO(model_bytesio, helper):
    """ Load a model from a BytesIO object.
    :param model_bytesio: A BytesIO object containing the model.
    :type model_bytesio: :class:`io.BytesIO`
    :param helper: The helper object for the model.
    :type helper: :class:`fedn.utils.helperbase.HelperBase`
    :return: The model object.
    :rtype: return type of helper.load
    """
    path = get_tmp_path()
    with open(path, 'wb') as fh:
        fh.write(model_bytesio)
        fh.flush()
    model = helper.load(path)
    os.unlink(path)
    return model


def serialize_model_to_BytesIO(model, helper):
    """ Serialize a model to a BytesIO object.

    :param model: The model object.
    :type model: return type of helper.load
    :param helper: The helper object for the model.
    :type helper: :class:`fedn.utils.helperbase.HelperBase`
    :return: A BytesIO object containing the model.
    :rtype: :class:`io.BytesIO`
    """
    outfile_name = helper.save(model)

    a = BytesIO()
    a.seek(0, 0)
    with open(outfile_name, 'rb') as f:
        a.write(f.read())
    a.seek(0)
    os.unlink(outfile_name)
    return a


class ModelService(rpc.ModelServiceServicer):
    """ Service for handling download and upload of models to the server.

    """

    def __init__(self):
        self.temp_model_storage = TempModelStorage()

    def exist(self, model_id):
        """ Check if a model exists on the server.

        :param model_id: The model id.
        :return: True if the model exists, else False.
        """
        return self.temp_model_storage.exist(model_id)

    def get_model(self, id):
        """ Download model with id 'id' from server.

        :param id: The model id.
        :type id: str
        :return: A BytesIO object containing the model.
        :rtype: :class:`io.BytesIO`, None if model does not exist.
        """

        data = BytesIO()
        data.seek(0, 0)

        parts = self.Download(fedn.ModelRequest(id=id), self)
        for part in parts:
            if part.status == fedn.ModelStatus.IN_PROGRESS:
                data.write(part.data)

            if part.status == fedn.ModelStatus.OK:
                return data
            if part.status == fedn.ModelStatus.FAILED:
                return None

    def set_model(self, model, id):
        """ Upload model to server.

        :param model: A model object (BytesIO)
        :type model: :class:`io.BytesIO`
        :param id: The model id.
        :type id: str
        """
        bt = model_as_bytesIO(model)
        # TODO: Check result
        _ = self.Upload(upload_request_generator(bt, id), self)

    # Model Service
    def Upload(self, request_iterator, context):
        """ RPC endpoints for uploading a model.

        :param request_iterator: The model request iterator.
        :type request_iterator: :class:`fedn.network.grpc.fedn_pb2.ModelRequest`
        :param context: The context object (unused)
        :type context: :class:`grpc._server._Context`
        :return: A model response object.
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ModelResponse`
        """
        logger.debug("grpc.ModelService.Upload: Called")
        result = None
        for request in request_iterator:
            if request.status == fedn.ModelStatus.IN_PROGRESS:
                self.temp_model_storage.get_ptr(request.id).write(request.data)
                self.temp_model_storage.set_model_metadata(request.id, fedn.ModelStatus.IN_PROGRESS)

            if request.status == fedn.ModelStatus.OK and not request.data:
                result = fedn.ModelResponse(id=request.id, status=fedn.ModelStatus.OK,
                                            message="Got model successfully.")
                # self.temp_model_storage_metadata.update({request.id: fedn.ModelStatus.OK})
                self.temp_model_storage.set_model_metadata(request.id, fedn.ModelStatus.OK)
                self.temp_model_storage.get_ptr(request.id).flush()
                self.temp_model_storage.get_ptr(request.id).close()
                return result

    def Download(self, request, context):
        """ RPC endpoints for downloading a model.

        :param request: The model request object.
        :type request: :class:`fedn.network.grpc.fedn_pb2.ModelRequest`
        :param context: The context object (unused)
        :type context: :class:`grpc._server._Context`
        :return: A model response iterator.
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ModelResponse`
        """
        logger.info(f'grpc.ModelService.Download: {request.sender.role}:{request.sender.name} requested model {request.id}')
        try:
            status = self.temp_model_storage.get_model_metadata(request.id)
            if status != fedn.ModelStatus.OK:
                logger.error(f'model file is not ready: {request.id}, status: {status}')
                yield fedn.ModelResponse(id=request.id, data=None, status=status)
        except Exception:
            logger.error("Error file does not exist: {}".format(request.id))
            yield fedn.ModelResponse(id=request.id, data=None, status=fedn.ModelStatus.FAILED)

        try:
            obj = self.temp_model_storage.get(request.id)
            if obj is None:
                raise Exception(f'File not found: {request.id}')
            with obj as f:
                while True:
                    piece = f.read(CHUNK_SIZE)
                    if len(piece) == 0:
                        yield fedn.ModelResponse(id=request.id, data=None, status=fedn.ModelStatus.OK)
                        return
                    yield fedn.ModelResponse(id=request.id, data=piece, status=fedn.ModelStatus.IN_PROGRESS)
        except Exception as e:
            logger.error("Downloading went wrong: {} {}".format(request.id, e))
            yield fedn.ModelResponse(id=request.id, data=None, status=fedn.ModelStatus.FAILED)



# This file contains the PackageRuntime class, which is used to download, validate and unpack compute packages.
#
#
import cgi
import os
import tarfile
from distutils.dir_util import copy_tree

import requests
import yaml

from fedn.common.log_config import logger
from fedn.utils.checksum import sha
from fedn.utils.dispatcher import Dispatcher


class PackageRuntime:
    """ PackageRuntime is used to download, validate and unpack compute packages.

    :param package_path: path to compute package
    :type package_path: str
    :param package_dir: directory to unpack compute package
    :type package_dir: str
    """

    def __init__(self, package_path, package_dir):

        self.dispatch_config = {'entry_points':
                                {'predict': {'command': 'python3 predict.py'},
                                 'train': {'command': 'python3 train.py'},
                                 'validate': {'command': 'python3 validate.py'}}}

        self.pkg_path = package_path
        self.pkg_name = None
        self.dir = package_dir
        self.checksum = None
        self.expected_checksum = None

    def download(self, host, port, token, force_ssl=False, secure=False, name=None):
        """ Download compute package from controller

        :param host: host of controller
        :param port: port of controller
        :param token: token for authentication
        :param name: name of package
        :return: True if download was successful, None otherwise
        :rtype: bool
        """
        # for https we assume a an ingress handles permanent redirect (308)
        if force_ssl:
            scheme = "https"
        else:
            scheme = "http"
        if port:
            path = f"{scheme}://{host}:{port}/download_package"
        else:
            path = f"{scheme}://{host}/download_package"
        if name:
            path = path + "?name={}".format(name)

        with requests.get(path, stream=True, verify=False, headers={'Authorization': 'Token {}'.format(token)}) as r:
            if 200 <= r.status_code < 204:

                params = cgi.parse_header(
                    r.headers.get('Content-Disposition', ''))[-1]
                try:
                    self.pkg_name = params['filename']
                except KeyError:
                    logger.error("No package returned.")
                    return None
                r.raise_for_status()
                with open(os.path.join(self.pkg_path, self.pkg_name), 'wb') as f:
                    for chunk in r.iter_content(chunk_size=8192):
                        f.write(chunk)
        if port:
            path = f"{scheme}://{host}:{port}/get_package_checksum"
        else:
            path = f"{scheme}://{host}/get_package_checksum"

        if name:
            path = path + "?name={}".format(name)
        with requests.get(path, verify=False, headers={'Authorization': 'Token {}'.format(token)}) as r:
            if 200 <= r.status_code < 204:

                data = r.json()
                try:
                    self.checksum = data['checksum']
                except Exception:
                    logger.error("Could not extract checksum.")

        return True

    def validate(self, expected_checksum):
        """ Validate the package against the checksum provided by the controller

        :param expected_checksum: checksum provided by the controller
        :return: True if checksums match, False otherwise
        :rtype: bool
        """
        self.expected_checksum = expected_checksum

        # crosscheck checksum and unpack if security checks are ok.
        file_checksum = str(sha(os.path.join(self.pkg_path, self.pkg_name)))

        if self.checksum == self.expected_checksum == file_checksum:
            logger.info("Package validated {}".format(self.checksum))
            return True
        else:
            return False

    def unpack(self):
        """ Unpack the compute package

        :return: True if unpacking was successful, False otherwise
        :rtype: bool
        """
        if self.pkg_name:
            f = None
            if self.pkg_name.endswith('tar.gz'):
                f = tarfile.open(os.path.join(
                    self.pkg_path, self.pkg_name), 'r:gz')
            if self.pkg_name.endswith('.tgz'):
                f = tarfile.open(os.path.join(
                    self.pkg_path, self.pkg_name), 'r:gz')
            if self.pkg_name.endswith('tar.bz2'):
                f = tarfile.open(os.path.join(
                    self.pkg_path, self.pkg_name), 'r:bz2')
        else:
            logger.error(
                "Failed to unpack compute package, no pkg_name set."
                "Has the reducer been configured with a compute package?"
            )
            return False

        os.getcwd()
        try:
            os.chdir(self.dir)

            if f:
                f.extractall()
                logger.info("Successfully extracted compute package content in {}".format(self.dir))
                return True
        except Exception:
            logger.error("Error extracting files.")
            return False

    def dispatcher(self, run_path):
        """ Dispatch the compute package

        :param run_path: path to dispatch the compute package
        :type run_path: str
        :return: Dispatcher object
        :rtype: :class:`fedn.utils.dispatcher.Dispatcher`
        """
        from_path = os.path.join(os.getcwd(), 'client')

        # preserve_times=False ensures compatibility with Gramine LibOS
        copy_tree(from_path, run_path, preserve_times=False)

        try:
            cfg = None
            with open(os.path.join(run_path, 'fedn.yaml'), 'rb') as config_file:

                cfg = yaml.safe_load(config_file.read())
                self.dispatch_config = cfg

        except Exception:
            logger.error(
                "Error trying to load and unpack dispatcher config - trying default")

        dispatcher = Dispatcher(self.dispatch_config, run_path)

        return dispatcher



import os

from .certificate import Certificate


class CertificateManager:
    """
    Utility to handle certificates for both Reducer and Combiner services.

    """

    def __init__(self, directory):

        self.directory = directory
        self.certificates = []
        self.allowed = dict()
        self.load_all()

    def get_or_create(self, name):
        """
        Look for an existing certificate, if not found, generate a self-signed certificate based on name.

        :param name: The name used when issuing the certificate.
        :return: A certificate
        :rtype: str
        """
        search = self.find(name)
        if search:
            return search
        else:
            cert = Certificate(self.directory, name=name,
                               cert_name=name + '-cert.pem', key_name=name + '-key.pem')
            cert.gen_keypair()
            self.certificates.append(cert)
            return cert

    def add(self, certificate):
        """
        Add certificate to certificate list.

        :param certificate:
        :return: Success status (True, False)
        :rtype: Boolean
        """
        if not self.find(certificate.name):
            self.certificates.append(certificate)
            return True
        return False

    def load_all(self):
        """
        Load all certificates and add to certificates list.

        """
        for filename in sorted(os.listdir(self.directory)):
            if filename.endswith('cert.pem'):
                name = filename.split('-')[0]
                key_name = name + '-key.pem'

                c = Certificate(self.directory, name=name,
                                cert_name=filename, key_name=key_name)
                self.certificates.append(c)

    def find(self, name):
        """

        :param name: Name of certificate
        :return: certificate if successful, else None
        """
        for cert in self.certificates:
            if cert.name == name:
                return cert
        for cert in self.allowed:
            if cert.name == name:
                return cert
        return None



import json
import queue
from abc import ABC, abstractmethod

import fedn.common.net.grpc.fedn_pb2 as fedn


class Aggregator(ABC):
    """ Abstract class defining an aggregator. """

    @abstractmethod
    def __init__(self, id, storage, server, modelservice, control):
        """ Initialize the aggregator.

        :param id: A reference to id of :class: `fedn.network.combiner.Combiner`
        :type id: str
        :param storage: Model repository for :class: `fedn.network.combiner.Combiner`
        :type storage: class: `fedn.common.storage.s3.s3repo.S3ModelRepository`
        :param server: A handle to the Combiner class :class: `fedn.network.combiner.Combiner`
        :type server: class: `fedn.network.combiner.Combiner`
        :param modelservice: A handle to the model service :class: `fedn.network.combiner.modelservice.ModelService`
        :type modelservice: class: `fedn.network.combiner.modelservice.ModelService`
        :param control: A handle to the :class: `fedn.network.combiner.round.RoundController`
        :type control: class: `fedn.network.combiner.round.RoundController`
        """
        self.name = self.__class__.__name__
        self.storage = storage
        self.id = id
        self.server = server
        self.modelservice = modelservice
        self.control = control
        self.model_updates = queue.Queue()

    @abstractmethod
    def combine_models(self, nr_expected_models=None, nr_required_models=1, helper=None, timeout=180):
        """Routine for combining model updates. Implemented in subclass.

        :param nr_expected_models: Number of expected models. If None, wait for all models.
        :type nr_expected_models: int
        :param nr_required_models: Number of required models to combine.
        :type nr_required_models: int
        :param helper: A helper object.
        :type helper: :class: `fedn.utils.plugins.helperbase.HelperBase`
        :param timeout: Timeout in seconds to wait for models to be combined.
        :type timeout: int
        :return: A combined model.
        """
        pass

    def on_model_update(self, model_update):
        """Callback when a new client model update is recieved.
           Performs (optional) pre-processing and then puts the update id
           on the aggregation queue. Override in subclass as needed.

        :param model_update: A ModelUpdate message.
        :type model_id: str
        """
        try:
            self.server.report_status("AGGREGATOR({}): callback received model update {}".format(self.name, model_update.model_update_id),
                                      log_level=fedn.Status.INFO)

            # Validate the update and metadata
            valid_update = self._validate_model_update(model_update)
            if valid_update:
                # Push the model update to the processing queue
                self.model_updates.put(model_update)
            else:
                self.server.report_status("AGGREGATOR({}): Invalid model update, skipping.".format(self.name))
        except Exception as e:
            self.server.report_status("AGGREGATOR({}): Failed to receive candidate model! {}".format(self.name, e),
                                      log_level=fedn.Status.WARNING)
            pass

    def on_model_validation(self, model_validation):
        """ Callback when a new client model validation is recieved.
            Performs (optional) pre-processing and then writes the validation
            to the database. Override in subclass as needed.

        :param validation: Dict containing validation data sent by client.
                           Must be valid JSON.
        :type validation: dict
        """

        # self.report_validation(validation)
        self.server.report_status("AGGREGATOR({}): callback processed validation {}".format(self.name, model_validation.model_id),
                                  log_level=fedn.Status.INFO)

    def _validate_model_update(self, model_update):
        """ Validate the model update.

        :param model_update: A ModelUpdate message.
        :type model_update: object
        :return: True if the model update is valid, False otherwise.
        :rtype: bool
        """
        # TODO: Validate the metadata to check that it contains all variables assumed by the aggregator.
        data = json.loads(model_update.meta)['training_metadata']
        if 'num_examples' not in data.keys():
            self.server.report_status("AGGREGATOR({}): Model validation failed, num_examples missing in metadata.".format(self.name))
            return False
        return True

    def next_model_update(self, helper):
        """ Get the next model update from the queue.

        :param helper: A helper object.
        :type helper: object
        :return: A tuple containing the model update, metadata and model id.
        :rtype: tuple
        """
        model_update = self.model_updates.get(block=False)
        model_id = model_update.model_update_id
        model_next = self.control.load_model_update(helper, model_id)
        # Get relevant metadata
        data = json.loads(model_update.meta)['training_metadata']
        config = json.loads(json.loads(model_update.meta)['config'])
        data['round_id'] = config['round_id']

        return model_next, data, model_id



from abc import ABC


class Config(ABC):
    pass


class ReducerConfig(Config):
    """ Configuration for the Reducer component. """
    compute_bundle_dir = None
    models_dir = None

    initial_model = None

    storage_backend = {
        'type': 's3', 'settings':
            {'bucket': 'models'}
    }

    def __init__(self):
        pass



""" The aggregator package is responsible for aggregating models from multiple clients. It's called both in :class:`fedn.network.combiner.Combiner` and :class:`fedn.network.controller.Controller` 
to aggregate models from clients. """
# flake8: noqa



import unittest

from fedn.network.clients.client import Client


class TestClient(unittest.TestCase):
    """Test the Client class."""

    def setUp(self):
        self.client = Client()

    def test_add_grpc_metadata(self):
        """Test the _add_grpc_metadata method."""

        # Test adding metadata when it doesn't exist
        self.client._add_grpc_metadata('key1', 'value1')
        self.assertEqual(self.client.metadata, (('key1', 'value1'),))

        # Test adding metadata when it already exists
        self.client._add_grpc_metadata('key1', 'value2')
        self.assertEqual(self.client.metadata, (('key1', 'value2'),))

        # Test adding multiple metadata
        self.client._add_grpc_metadata('key2', 'value3')
        self.assertEqual(self.client.metadata, (('key1', 'value2'), ('key2', 'value3')))

        # Test adding metadata with special characters
        self.client._add_grpc_metadata('key3', 'value4!@#$%^&*()')
        self.assertEqual(self.client.metadata, (('key1', 'value2'), ('key2', 'value3'), ('key3', 'value4!@#$%^&*()')))

        # Test adding metadata with empty key
        with self.assertRaises(ValueError):
            self.client._add_grpc_metadata('', 'value5')

        # Test adding metadata with empty value
        with self.assertRaises(ValueError):
            self.client._add_grpc_metadata('key4', '')

        # Test adding metadata with None value
        with self.assertRaises(ValueError):
            self.client._add_grpc_metadata('key5', None)


if __name__ == '__main__':
    unittest.main()



import logging
import logging.config

import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
logging.getLogger("urllib3").setLevel(logging.ERROR)

handler = logging.StreamHandler()
logger = logging.getLogger()
logger.addHandler(handler)
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s [%(levelname)s] %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
handler.setFormatter(formatter)


def set_log_level_from_string(level_str):
    """
    Set the log level based on a string input.
    """
    # Mapping of string representation to logging constants
    level_mapping = {
        'CRITICAL': logging.CRITICAL,
        'ERROR': logging.ERROR,
        'WARNING': logging.WARNING,
        'INFO': logging.INFO,
        'DEBUG': logging.DEBUG,
    }

    # Get the logging level from the mapping
    level = level_mapping.get(level_str.upper())

    if not level:
        raise ValueError(f"Invalid log level: {level_str}")

    # Set the log level
    logger.setLevel(level)


def set_log_stream(log_file):
    """
    Redirect the log stream to a specified file, if log_file is set.
    """
    if not log_file:
        return

    # Remove existing handlers
    for h in logger.handlers[:]:
        logger.removeHandler(h)

    # Create a FileHandler
    file_handler = logging.FileHandler(log_file)
    file_handler.setFormatter(formatter)

    # Add the file handler to the logger
    logger.addHandler(file_handler)



import copy
import os
import random
import uuid

from OpenSSL import crypto


class Certificate:
    """
    Utility to generate unsigned certificates.

    """
    CERT_NAME = "cert.pem"
    KEY_NAME = "key.pem"
    BITS = 2048

    def __init__(self, cwd, name=None, key_name="key.pem", cert_name="cert.pem", create_dirs=True):

        try:
            os.makedirs(cwd)
        except OSError:
            print("Directory exists, will store all cert and keys here.")
        else:
            print(
                "Successfully created the directory to store cert and keys in {}".format(cwd))

        self.key_path = os.path.join(cwd, key_name)
        self.cert_path = os.path.join(cwd, cert_name)

        if name:
            self.name = name
        else:
            self.name = str(uuid.uuid4())

    def gen_keypair(self, ):
        """
        Generate keypair.

        """
        key = crypto.PKey()
        key.generate_key(crypto.TYPE_RSA, 2048)
        cert = crypto.X509()
        cert.get_subject().C = "SE"
        cert.get_subject().ST = "Stockholm"
        cert.get_subject().O = "Development Key"  # noqa: E741
        cert.get_subject().OU = "Development Key"
        cert.get_subject().CN = self.name

        cert.set_serial_number(int(random.randint(1000, 100000)))

        cert.gmtime_adj_notBefore(0)
        cert.gmtime_adj_notAfter(31 * 24 * 60 * 60)

        cert.set_issuer(cert.get_subject())
        cert.set_pubkey(key)
        cert.sign(key, "sha256")
        with open(self.key_path, "wb") as keyfile:
            keyfile.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, key))

        with open(self.cert_path, "wb") as certfile:
            certfile.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))

    def set_keypair_raw(self, certificate, privatekey):
        """

        :param certificate:
        :param privatekey:
        """
        with open(self.key_path, "wb") as keyfile:
            keyfile.write(crypto.dump_privatekey(
                crypto.FILETYPE_PEM, privatekey))

        with open(self.cert_path, "wb") as certfile:
            certfile.write(crypto.dump_certificate(
                crypto.FILETYPE_PEM, certificate))

    def get_keypair_raw(self):
        """

        :return:
        """
        with open(self.key_path, 'rb') as keyfile:
            key_buf = keyfile.read()
        with open(self.cert_path, 'rb') as certfile:
            cert_buf = certfile.read()
        return copy.deepcopy(cert_buf), copy.deepcopy(key_buf)

    def get_key(self):
        """

        :return:
        """
        with open(self.key_path, 'rb') as keyfile:
            key_buf = keyfile.read()
        key = crypto.load_privatekey(crypto.FILETYPE_PEM, key_buf)
        return key

    def get_cert(self):
        """

        :return:
        """
        with open(self.cert_path, 'rb') as certfile:
            cert_buf = certfile.read()
        cert = crypto.load_certificate(crypto.FILETYPE_PEM, cert_buf)
        return cert

    def __str__(self):
        return "Certificate name: {}".format(self.name)



import importlib
import json
import queue
from abc import ABC, abstractmethod

from fedn.common.log_config import logger

AGGREGATOR_PLUGIN_PATH = "fedn.network.combiner.aggregators.{}"


class AggregatorBase(ABC):
    """ Abstract class defining an aggregator.

    :param id: A reference to id of :class: `fedn.network.combiner.Combiner`
    :type id: str
    :param storage: Model repository for :class: `fedn.network.combiner.Combiner`
    :type storage: class: `fedn.common.storage.s3.s3repo.S3ModelRepository`
    :param server: A handle to the Combiner class :class: `fedn.network.combiner.Combiner`
    :type server: class: `fedn.network.combiner.Combiner`
    :param modelservice: A handle to the model service :class: `fedn.network.combiner.modelservice.ModelService`
    :type modelservice: class: `fedn.network.combiner.modelservice.ModelService`
    :param control: A handle to the :class: `fedn.network.combiner.roundhandler.RoundHandler`
    :type control: class: `fedn.network.combiner.roundhandler.RoundHandler`
    """

    @abstractmethod
    def __init__(self, storage, server, modelservice, round_handler):
        """ Initialize the aggregator."""
        self.name = self.__class__.__name__
        self.storage = storage
        self.server = server
        self.modelservice = modelservice
        self.round_handler = round_handler
        self.model_updates = queue.Queue()

    @abstractmethod
    def combine_models(self, nr_expected_models=None, nr_required_models=1, helper=None, timeout=180, delete_models=True):
        """Routine for combining model updates. Implemented in subclass.

        :param nr_expected_models: Number of expected models. If None, wait for all models.
        :type nr_expected_models: int
        :param nr_required_models: Number of required models to combine.
        :type nr_required_models: int
        :param helper: A helper object.
        :type helper: :class: `fedn.utils.plugins.helperbase.HelperBase`
        :param timeout: Timeout in seconds to wait for models to be combined.
        :type timeout: int
        :param delete_models: Delete client models after combining.
        :type delete_models: bool
        :return: A combined model.
        """
        pass

    def on_model_update(self, model_update):
        """Callback when a new client model update is recieved.

        Performs (optional) validation and pre-processing,
        and then puts the update id on the aggregation queue.
        Override in subclass as needed.

        :param model_update: fedn.network.grpc.fedn.proto.ModelUpdate
        :type model_id: str
        """
        try:
            logger.info("AGGREGATOR({}): callback received model update {}".format(self.name, model_update.model_update_id))

            # Validate the update and metadata
            valid_update = self._validate_model_update(model_update)
            if valid_update:
                # Push the model update to the processing queue
                self.model_updates.put(model_update)
            else:
                logger.warning("AGGREGATOR({}): Invalid model update, skipping.".format(self.name))
        except Exception as e:
            logger.error("AGGREGATOR({}): failed to receive model update! {}".format(self.name, e))
            pass

    def _validate_model_update(self, model_update):
        """ Validate the model update.

        :param model_update: A ModelUpdate message.
        :type model_update: object
        :return: True if the model update is valid, False otherwise.
        :rtype: bool
        """
        data = json.loads(model_update.meta)['training_metadata']
        if 'num_examples' not in data.keys():
            logger.error("AGGREGATOR({}): Model validation failed, num_examples missing in metadata.".format(self.name))
            return False
        return True

    def next_model_update(self):
        """ Get the next model update from the queue.

        :param helper: A helper object.
        :type helper: object
        :return: The model update.
        :rtype: fedn.network.grpc.fedn.proto.ModelUpdate
        """
        model_update = self.model_updates.get(block=False)
        return model_update

    def load_model_update(self, model_update, helper):
        """ Load the memory representation of the model update.

        Load the model update paramters and the
        associate metadata into memory.

        :param model_update: The model update.
        :type model_update: fedn.network.grpc.fedn.proto.ModelUpdate
        :param helper: A helper object.
        :type helper: fedn.utils.helpers.helperbase.Helper
        :return: A tuple of (parameters, metadata)
        :rtype: tuple
        """
        model_id = model_update.model_update_id
        model = self.round_handler.load_model_update(helper, model_id)
        # Get relevant metadata
        data = json.loads(model_update.meta)['training_metadata']
        config = json.loads(json.loads(model_update.meta)['config'])
        data['round_id'] = config['round_id']

        return model, data

    def get_state(self):
        """ Get the state of the aggregator's queue, including the number of model updates."""
        state = {
            'queue_len': self.model_updates.qsize()
        }
        return state


def get_aggregator(aggregator_module_name, storage, server, modelservice, control):
    """ Return an instance of the helper class.

    :param helper_module_name: The name of the helper plugin module.
    :type helper_module_name: str
    :param storage: Model repository for :class: `fedn.network.combiner.Combiner`
    :type storage: class: `fedn.common.storage.s3.s3repo.S3ModelRepository`
    :param server: A handle to the Combiner class :class: `fedn.network.combiner.Combiner`
    :type server: class: `fedn.network.combiner.Combiner`
    :param modelservice: A handle to the model service :class: `fedn.network.combiner.modelservice.ModelService`
    :type modelservice: class: `fedn.network.combiner.modelservice.ModelService`
    :param control: A handle to the :class: `fedn.network.combiner.roundhandler.RoundHandler`
    :type control: class: `fedn.network.combiner.roundhandler.RoundHandler`
    :return: An aggregator instance.
    :rtype: class: `fedn.combiner.aggregators.AggregatorBase`
    """
    aggregator_plugin = AGGREGATOR_PLUGIN_PATH.format(aggregator_module_name)
    aggregator = importlib.import_module(aggregator_plugin)
    return aggregator.Aggregator(storage, server, modelservice, control)




import numpy as np

from fedn.utils.helpers.helperbase import HelperBase


class Helper(HelperBase):
    """ FEDn helper class for models weights/parameters that can be transformed to numpy ndarrays. """

    def __init__(self):
        """ Initialize helper. """
        super().__init__()
        self.name = "numpyhelper"

    def increment_average(self, m1, m2, n, N):
        """ Update a weighted incremental average of model weights.

        :param m1: Current parameters.
        :type model: list of numpy ndarray
        :param m2: next parameters.
        :type model_next: list of numpy ndarray
        :param n: Number of examples used for updating m2.
        :type n: int
        :param N: Total number of examples (accumulated).
        :type N: int
        :return: Updated incremental weighted average.
        :rtype: list of numpy ndarray
        """

        return [np.add(x, n*(y-x)/N) for x, y in zip(m1, m2)]

    def add(self, m1, m2, a=1.0, b=1.0):
        """ m1*a + m2*b

        :param model: Current model weights.
        :type model: list of ndarrays
        :param model_next: New model weights.
        :type model_next: list of ndarrays
        :return: Incremental weighted average of model weights.
        :rtype: list of ndarrays
        """

        return [x*a+y*b for x, y in zip(m1, m2)]

    def subtract(self, m1, m2, a=1.0, b=1.0):
        """ m1*a - m2*b.

        :param m1: Current model weights.
        :type m1: list of ndarrays
        :param m2: New model weights.
        :type m2: list of ndarrays
        :return: m1*a-m2*b
        :rtype: list of ndarrays
        """
        return self.add(m1, m2, a, -b)

    def divide(self, m1, m2):
        """ Subtract weights.

        :param m1: Current model weights.
        :type m1: list of ndarrays
        :param m2: New model weights.
        :type m2: list of ndarrays
        :return: m1/m2.
        :rtype: list of ndarrays
        """

        return [np.divide(x, y) for x, y in zip(m1, m2)]

    def multiply(self, m1, m2):
        """ Multiply m1 by m2.

        :param m1: Current model weights.
        :type m1: list of ndarrays
        :param m2: New model weights.
        :type m2: list of ndarrays
        :return: m1.*m2
        :rtype: list of ndarrays
        """

        return [np.multiply(x, y) for (x, y) in zip(m1, m2)]

    def sqrt(self, m1):
        """ Sqrt of m1, element-wise.

        :param m1: Current model weights.
        :type model: list of ndarrays
        :param model_next: New model weights.
        :type model_next: list of ndarrays
        :return: sqrt(m1)
        :rtype: list of ndarrays
        """

        return [np.sqrt(x) for x in m1]

    def power(self, m1, a):
        """ m1 raised to the power of m2.

        :param m1: Current model weights.
        :type m1: list of ndarrays
        :param m2: New model weights.
        :type a: float
        :return: m1.^m2
        :rtype: list of ndarrays
        """

        return [np.power(x, a) for x in m1]

    def norm(self, m):
        """ Return the norm (L1) of model weights.

        :param m: Current model weights.
        :type m: list of ndarrays
        :return: norm of m
        :rtype: float
        """
        n = 0.0
        for x in m:
            n += np.linalg.norm(x, 1)
        return n

    def ones(self, m1, a):
        """ Return a list of numpy arrays of the same shape as m1, filled with ones.

        :param m1: Current model weights.
        :type m1: list of ndarrays
        :param a: Scalar value.
        :type a: float
        :return: list of numpy arrays of the same shape as m1, filled with ones.
        :rtype: list of ndarrays
        """

        res = []
        for x in m1:
            res.append(np.ones(np.shape(x))*a)
        return res

    def save(self, weights, path=None):
        """ Serialize weights to file. The serialized model must be a single binary object.

        :param weights: List of weights in numpy format.
        :param path: Path to file.
        :return: Path to file.
        """
        if not path:
            path = self.get_tmp_path()

        weights_dict = {}
        for i, w in enumerate(weights):
            weights_dict[str(i)] = w

        np.savez_compressed(path, **weights_dict)

        return path

    def load(self, fh):
        """ Load weights from file or filelike.

        :param fh: file path, filehandle, filelike.
        :return: List of weights in numpy format.
        """
        a = np.load(fh)

        weights = []
        for i in range(len(a.files)):
            weights.append(a[str(i)])
        return weights



import uuid

import click
import yaml

from fedn.common.exceptions import InvalidClientConfig
from fedn.common.log_config import logger
from fedn.network.clients.client import Client
from fedn.network.combiner.combiner import Combiner

from .main import main


def get_statestore_config_from_file(init):
    """

    :param init:
    :return:
    """
    with open(init, 'r') as file:
        try:
            settings = dict(yaml.safe_load(file))
            return settings
        except yaml.YAMLError as e:
            raise (e)


def check_helper_config_file(config):
    control = config['control']
    try:
        helper = control["helper"]
    except KeyError:
        logger.error("--local-package was used, but no helper was found in --init settings file.")
        exit(-1)
    return helper


def apply_config(config):
    """Parse client config from file.

    Override configs from the CLI with settings in config file.

    :param config: Client config (dict).
    """
    with open(config['init'], 'r') as file:
        try:
            settings = dict(yaml.safe_load(file))
        except Exception:
            logger.error('Failed to read config from settings file, exiting.')
            return

    for key, val in settings.items():
        config[key] = val


def validate_client_config(config):
    """Validate client configuration.

    :param config: Client config (dict).
    """

    try:
        if config['discover_host'] is None or \
                config['discover_host'] == '':
            raise InvalidClientConfig("Missing required configuration: discover_host")
        if 'discover_port' not in config.keys():
            config['discover_port'] = None
    except Exception:
        raise InvalidClientConfig("Could not load config from file. Check config")


@main.group('run')
@click.pass_context
def run_cmd(ctx):
    """

    :param ctx:
    """
    pass


@run_cmd.command('client')
@click.option('-d', '--discoverhost', required=False, help='Hostname for discovery services(reducer).')
@click.option('-p', '--discoverport', required=False, help='Port for discovery services (reducer).')
@click.option('--token', required=False, help='Set token provided by reducer if enabled')
@click.option('-n', '--name', required=False, default="client" + str(uuid.uuid4())[:8])
@click.option('-i', '--client_id', required=False)
@click.option('--local-package', is_flag=True, help='Enable local compute package')
@click.option('--force-ssl', is_flag=True, help='Force SSL/TLS for REST service')
@click.option('-u', '--dry-run', required=False, default=False)
@click.option('-s', '--secure', required=False, default=False)
@click.option('-pc', '--preshared-cert', required=False, default=False)
@click.option('-v', '--verify', is_flag=True, help='Verify SSL/TLS for REST service')
@click.option('-c', '--preferred-combiner', required=False, default=False)
@click.option('-va', '--validator', required=False, default=True)
@click.option('-tr', '--trainer', required=False, default=True)
@click.option('-in', '--init', required=False, default=None,
              help='Set to a filename to (re)init client from file state.')
@click.option('-l', '--logfile', required=False, default=None,
              help='Set logfile for client log to file.')
@click.option('--heartbeat-interval', required=False, default=2)
@click.option('--reconnect-after-missed-heartbeat', required=False, default=30)
@click.option('--verbosity', required=False, default='INFO', type=click.Choice(['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'], case_sensitive=False))
@click.pass_context
def client_cmd(ctx, discoverhost, discoverport, token, name, client_id, local_package, force_ssl, dry_run, secure, preshared_cert,
               verify, preferred_combiner, validator, trainer, init, logfile, heartbeat_interval, reconnect_after_missed_heartbeat,
               verbosity):
    """

    :param ctx:
    :param discoverhost:
    :param discoverport:
    :param token:
    :param name:
    :param client_id:
    :param remote:
    :param dry_run:
    :param secure:
    :param preshared_cert:
    :param verify_cert:
    :param preferred_combiner:
    :param init:
    :param logfile:
    :param hearbeat_interval
    :param reconnect_after_missed_heartbeat
    :param verbosity
    :return:
    """
    remote = False if local_package else True
    config = {'discover_host': discoverhost, 'discover_port': discoverport, 'token': token, 'name': name,
              'client_id': client_id, 'remote_compute_context': remote, 'force_ssl': force_ssl, 'dry_run': dry_run, 'secure': secure,
              'preshared_cert': preshared_cert, 'verify': verify, 'preferred_combiner': preferred_combiner,
              'validator': validator, 'trainer': trainer, 'init': init, 'logfile': logfile, 'heartbeat_interval': heartbeat_interval,
              'reconnect_after_missed_heartbeat': reconnect_after_missed_heartbeat, 'verbosity': verbosity}

    if init:
        apply_config(config)

    validate_client_config(config)

    client = Client(config)
    client.run()


@run_cmd.command('combiner')
@click.option('-d', '--discoverhost', required=False, help='Hostname for discovery services (reducer).')
@click.option('-p', '--discoverport', required=False, help='Port for discovery services (reducer).')
@click.option('-t', '--token', required=False, help='Set token provided by reducer if enabled')
@click.option('-n', '--name', required=False, default="combiner" + str(uuid.uuid4())[:8], help='Set name for combiner.')
@click.option('-h', '--host', required=False, default="combiner", help='Set hostname.')
@click.option('-i', '--port', required=False, default=12080, help='Set port.')
@click.option('-f', '--fqdn', required=False, default=None, help='Set fully qualified domain name')
@click.option('-s', '--secure', is_flag=True, help='Enable SSL/TLS encrypted gRPC channels.')
@click.option('-v', '--verify', is_flag=True, help='Verify SSL/TLS for REST discovery service (reducer)')
@click.option('-c', '--max_clients', required=False, default=30, help='The maximal number of client connections allowed.')
@click.option('-in', '--init', required=False, default=None,
              help='Path to configuration file to (re)init combiner.')
@click.pass_context
def combiner_cmd(ctx, discoverhost, discoverport, token, name, host, port, fqdn, secure, verify, max_clients, init):
    """

    :param ctx:
    :param discoverhost:
    :param discoverport:
    :param token:
    :param name:
    :param hostname:
    :param port:
    :param secure:
    :param max_clients:
    :param init:
    """
    config = {'discover_host': discoverhost, 'discover_port': discoverport, 'token': token, 'host': host,
              'port': port, 'fqdn': fqdn, 'name': name, 'secure': secure, 'verify': verify, 'max_clients': max_clients,
              'init': init}

    if config['init']:
        apply_config(config)

    combiner = Combiner(config)
    combiner.run()



import importlib
import json

HELPER_PLUGIN_PATH = "fedn.utils.helpers.plugins.{}"


def get_helper(helper_module_name):
    """ Return an instance of the helper class.

    :param helper_module_name: The name of the helper plugin module.
    :type helper_module_name: str
    :return: A helper instance.
    :rtype: class: `fedn.utils.helpers.helpers.HelperBase`
    """
    helper_plugin = HELPER_PLUGIN_PATH.format(helper_module_name)
    helper = importlib.import_module(helper_plugin)
    return helper.Helper()


def save_metadata(metadata, filename):
    """ Save metadata to file.

    :param metadata: The metadata to save.
    :type metadata: dict
    :param filename: The name of the file to save to.
    :type filename: str
    """
    with open(filename+'-metadata', 'w') as outfile:
        json.dump(metadata, outfile)


def save_metrics(metrics, filename):
    """ Save metrics to file.

    :param metrics: The metrics to save.
    :type metrics: dict
    :param filename: The name of the file to save to.
    :type filename: str
    """
    with open(filename, 'w') as outfile:
        json.dump(metrics, outfile)



from concurrent import futures

import grpc
from grpc_health.v1 import health, health_pb2_grpc

import fedn.network.grpc.fedn_pb2_grpc as rpc
from fedn.common.log_config import (logger, set_log_level_from_string,
                                    set_log_stream)


class Server:
    """ Class for configuring and launching the gRPC server."""

    def __init__(self, servicer, modelservicer, config):

        set_log_level_from_string(config.get('verbosity', "INFO"))
        set_log_stream(config.get('logfile', None))

        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=350))
        self.certificate = None
        self.health_servicer = health.HealthServicer()

        if isinstance(servicer, rpc.CombinerServicer):
            rpc.add_CombinerServicer_to_server(servicer, self.server)
        if isinstance(servicer, rpc.ConnectorServicer):
            rpc.add_ConnectorServicer_to_server(servicer, self.server)
        if isinstance(servicer, rpc.ReducerServicer):
            rpc.add_ReducerServicer_to_server(servicer, self.server)
        if isinstance(modelservicer, rpc.ModelServiceServicer):
            rpc.add_ModelServiceServicer_to_server(modelservicer, self.server)
        if isinstance(servicer, rpc.CombinerServicer):
            rpc.add_ControlServicer_to_server(servicer, self.server)

        health_pb2_grpc.add_HealthServicer_to_server(self.health_servicer, self.server)

        if config['secure']:
            logger.info(f'Creating secure gRPCS server using certificate: {config["certificate"]}')
            server_credentials = grpc.ssl_server_credentials(
                ((config['key'], config['certificate'],),))
            self.server.add_secure_port(
                '[::]:' + str(config['port']), server_credentials)
        else:
            logger.info("Creating insecure gRPC server")
            self.server.add_insecure_port('[::]:' + str(config['port']))

    def start(self):
        """ Start the gRPC server."""
        logger.info("gRPC Server started")
        self.server.start()

    def stop(self):
        """ Stop the gRPC server."""
        logger.info("gRPC Server stopped")
        self.server.stop(0)



import os

import yaml

global STATESTORE_CONFIG
global MODELSTORAGE_CONFIG


def get_environment_config():
    """ Get the configuration from environment variables.
    """
    global STATESTORE_CONFIG
    global MODELSTORAGE_CONFIG

    STATESTORE_CONFIG = os.environ.get('STATESTORE_CONFIG',
                                       '/workspaces/fedn/config/settings-reducer.yaml.template')
    MODELSTORAGE_CONFIG = os.environ.get('MODELSTORAGE_CONFIG',
                                         '/workspaces/fedn/config/settings-reducer.yaml.template')


def get_statestore_config(file=None):
    """ Get the statestore configuration from file.

    :param file: The statestore configuration file (yaml) path (optional).
    :type file: str
    :return: The statestore configuration as a dict.
    :rtype: dict
    """
    if file is None:
        get_environment_config()
        file = STATESTORE_CONFIG
    with open(file, 'r') as config_file:
        try:
            settings = dict(yaml.safe_load(config_file))
        except yaml.YAMLError as e:
            raise (e)
    return settings["statestore"]


def get_modelstorage_config(file=None):
    """ Get the model storage configuration from file.

    :param file: The model storage configuration file (yaml) path (optional).
    :type file: str
    :return: The model storage configuration as a dict.
    :rtype: dict
    """
    if file is None:
        get_environment_config()
        file = MODELSTORAGE_CONFIG
    with open(file, 'r') as config_file:
        try:
            settings = dict(yaml.safe_load(config_file))
        except yaml.YAMLError as e:
            raise (e)
    return settings["storage"]


def get_network_config(file=None):
    """ Get the network configuration from file.

    :param file: The network configuration file (yaml) path (optional).
    :type file: str
    :return: The network id.
    :rtype: str
    """
    if file is None:
        get_environment_config()
        file = STATESTORE_CONFIG
    with open(file, 'r') as config_file:
        try:
            settings = dict(yaml.safe_load(config_file))
        except yaml.YAMLError as e:
            raise (e)
    return settings["network_id"]


def get_controller_config(file=None):
    """ Get the controller configuration from file.

    :param file: The controller configuration file (yaml) path (optional).
    :type file: str
    :return: The controller configuration as a dict.
    :rtype: dict
    """
    if file is None:
        get_environment_config()
        file = STATESTORE_CONFIG
    with open(file, 'r') as config_file:
        try:
            settings = dict(yaml.safe_load(config_file))
        except yaml.YAMLError as e:
            raise (e)
    return settings["controller"]
